
# TODO: figure out this
raise Exception("Deprecated?")

import urllib

import mako.filters


def u_path_escape(text):
    return urllib.parse.quote(text)


def install():

    mako.filters.DEFAULT_ESCAPES['u_path'] = 'filters.u_path_escape'
    mako.filters.u_path_escape = u_path_escape




"""
This module is for creating variable-like objects for storring size python
objects
"""

import os.path
import yaml

import wayround_org.utils.path


class PersistentMemoryDriver:
    pass


class PersistentMemoryDriverFileSystem(PersistentMemoryDriver):

    def __init__(self, directory):
        self.directory = directory
        self.counter = 0
        return

    def init(self):
        os.makedirs(self.directory, exist_ok=True)
        files = os.listdir(self.directory)
        for i in files:

            if not i.endswith('.yaml'):
                continue

            descriptor = int(i[0:i.find('.')])
            os.unlink(self.gen_descriptor_filepath(descriptor))
        return

    def gen_descriptor_filepath(self, descriptor):
        if not isinstance(descriptor, int):
            raise TypeError("`descriptor' must be int")
        if descriptor < 0:
            raise ValueError("`descriptor' must be >= 0")
        ret = wayround_org.utils.path.join(
            self.directory,
            '{}.yaml'.format(str(descriptor))
            )
        return ret

    def new(self):
        ret = self.counter
        self.counter += 1
        return ret

    def delete(self, descriptor):
        f_path = self.gen_descriptor_filepath(descriptor)
        if os.path.isfile(f_path):
            os.unlink(f_path)
        return

    def get_value(self, descriptor):
        ret = None
        f_path = self.gen_descriptor_filepath(descriptor)
        if os.path.isfile(f_path):
            with open(f_path) as f:
                ret = yaml.load(f.read())
        return ret

    def set_value(self, descriptor, value):
        f_path = self.gen_descriptor_filepath(descriptor)
        with open(f_path, 'w') as f:
            f.write(yaml.dump(value))
        return

    def open(self, descriptor, flags):
        f_path = self.gen_descriptor_filepath(descriptor)
        ret = open(f_path, flags)
        return ret

    def get_size(self):
        return os.stat(self.gen_descriptor_filepath(descriptor)).st_size

    def get_bytes(self, descriptor, size_limit=1024, bs=500):
        ret = None

        f_path = self.gen_descriptor_filepath(descriptor)

        if self.get_size(descriptor) <= size_limit:
            if os.path.isfile(f_path):
                ret = b''
                with open(f_path, 'br') as f:
                    while True:
                        b = f.read(bs)
                        if len(b) == 0:
                            break
                        ret += b

        return ret


class PersistentMemory:

    @classmethod
    def new_fs_memory(cls, path):
        return cls(PersistentMemoryDriverFileSystem(path))

    def __init__(self, driver):
        if not isinstance(driver, PersistentMemoryDriver):
            raise TypeError(
                "invalid `driver' instance type"
                )
        self.driver = driver
        return

    def new(self, value=None):
        return PersistentVariable(self.driver, value)

    def init(self):
        return self.driver.init()


class PersistentVariable:

    def __init__(self, driver, initial):
        if not isinstance(driver, PersistentMemoryDriver):
            raise TypeError(
                "invalid `driver' instance type"
                )
        self.driver = driver
        self.descriptor = self.driver.new()
        self.set(initial)
        return

    def get(self):
        return self.driver.get_value(self.descriptor)

    def set(self, value):
        return self.driver.set_value(self.descriptor, value)

    def __del__(self):
        self.driver.delete(self.descriptor)
        return

    def open(self, flags='r'):
        return self.driver.open(self.descriptor, flags)

    def get_size(self):
        return self.driver.get_size(self.descriptor)

    def get_bytes(self, size_limit=1024, bs=500):
        return self.driver.get_bytes(
            self,
            self.descriptor,
            size_limit=size_limit,
            bs=bs
            )





def convert_cb_params_to_boolean(params, names):

    for i in names:

        if i in params and (
            params[i].lower() in ['on', 'yes', 'true', 'ok']
            ):
            params[i] = True
        else:
            params[i] = False




import fnmatch
import logging
import os
import select
import shutil
import threading
import time
import copy

import wayround_org.utils.path
import wayround_org.utils.terminal
import wayround_org.utils.text


POLL_CONSTS = {}

_l = dir(select)
for _i in [
        'POLLIN',
        'POLLPRI',
        'POLLOUT',
        'POLLERR',
        'POLLHUP',
        'POLLNVAL'
        ]:
    POLL_CONSTS[_i] = eval('select.{}'.format(_i))

del _i
del _l


def copy_file(src, dst, overwrite_dst=False, verbose=True):

    ret = 0

    if os.path.isfile(dst) and not overwrite_dst:
        if verbose:
            logging.info("Skipping overwriting file `{}'".format(dst))
    else:

        if verbose:
            logging.info("copying `{}'".format(os.path.relpath(src)))

        try:
            shutil.copy2(src, dst)
        except:
            if verbose:
                logging.error(
                    "Can't overwrite file `{}'".format(dst)
                    )
            ret = 1

    return ret


def copy_symlink(src, dst, overwrite_dst=False, verbose=True):

    ret = 0

    link_value = os.readlink(src)

    if ((os.path.isfile(dst)
         or os.path.islink(dst))
            and overwrite_dst):

        os.unlink(dst)

    elif os.path.isdir(dst):
        if verbose:
            logging.error(
                "Can't create link. It's a directory `{}'".format(dst))
        ret = 1

    try:
        os.symlink(link_value, dst)
    except:
        if verbose:
            logging.error(
                "Can't create link. File exists `{}'".format(dst)
                )
        ret = 2

    return ret


def _copytree_sub(
        src_dir,
        dst_dir,
        overwrite_files=False,
        stop_on_overwrite_error=True,
        verbose=True
        ):

    # TODO: make verbose parameter

    ret = 0

    full_src_dir = wayround_org.utils.path.abspath(src_dir)
    full_dst_dir = wayround_org.utils.path.abspath(dst_dir)

    if not os.path.isdir(full_src_dir):
        if verbose:
            logging.error("Source dir not exists `{}'".format(src_dir))
        ret = 1
    else:
        if create_if_not_exists_dir(full_dst_dir) != 0:
            if verbose:
                logging.error(
                    "Error creating destination dir `{}'".format(full_dst_dir)
                    )
            ret = 2
        else:

            for path, dirs, files in os.walk(
                    src_dir,
                    topdown=True,
                    onerror=_copytree_on_error,
                    followlinks=False
                    ):

                path_dst = wayround_org.utils.path.join(
                    full_dst_dir,
                    wayround_org.utils.path.relpath(
                        path,
                        full_src_dir
                        )
                    )

                dirs.sort()
                files.sort()

                for i in dirs:

                    joined = wayround_org.utils.path.join(path, i)
                    joined_dst = wayround_org.utils.path.join(path_dst, i)

                    if os.path.islink(joined):

                        if copy_symlink(
                                joined,
                                joined_dst,
                                overwrite_files,
                                verbose=verbose
                                ) != 0:
                            ret = 3
                            # TODO: break?

                    else:
                        if create_if_not_exists_dir(joined_dst) != 0:
                            if verbose:
                                logging.error(
                                    "Can't create directory `{}'".format(
                                        joined_dst
                                        )
                                    )
                            ret = 5
                            # TODO: break?

                # TODO: if ret == 0?
                for i in files:

                    joined = wayround_org.utils.path.join(path, i)
                    joined_dst = wayround_org.utils.path.join(path_dst, i)

                    if os.path.islink(joined):

                        if copy_symlink(
                                joined,
                                joined_dst,
                                overwrite_files,
                                verbose=verbose
                                ) != 0:
                            ret = 3
                            break

                    elif os.path.isfile(joined):

                        if copy_file(
                                joined,
                                joined_dst,
                                overwrite_files,
                                verbose=verbose
                                ) != 0:
                            ret = 4
                            break

                if ret != 0:
                    break

    return ret


def _copytree_on_error(err):
    try:
        raise err
    except:
        logging.exception("Can't copy file:\n{}".format(err))
    return


def copytree(
        src_dir,
        dst_dir,
        overwrite_files=False,
        clear_before_copy=False,
        dst_must_be_empty=True,
        verbose=True
        ):

    # TODO: think of hardlinks too..

    src_dir = wayround_org.utils.path.abspath(src_dir)
    dst_dir = wayround_org.utils.path.abspath(dst_dir)

    ret = 0

    if verbose:
        logging.info("copying `{}'\n    to\n    `{}'".format(src_dir, dst_dir))

    if not os.path.exists(dst_dir):
        os.makedirs(dst_dir, exist_ok=True)

    if dst_must_be_empty:
        if not isdirempty(dst_dir):
            if verbose:
                logging.error("Destination dir `{}' not empty".format(dst_dir))
            ret = 1

    if ret == 0:

        if clear_before_copy:
            if verbose:
                logging.info("Cleaning dir `{}'".format(dst_dir))
            cleanup_dir(dst_dir, verbose=verbose)

        if create_if_not_exists_dir(dst_dir, verbose=verbose) != 0:
            if verbose:
                logging.error("Error creating dir `{}'".format(dst_dir))
            ret = 2
        else:
            if (_copytree_sub(
                    src_dir,
                    dst_dir,
                    overwrite_files=overwrite_files,
                    verbose=verbose
                    )
                    != 0):
                if verbose:
                    logging.error(
                        "Some errors occurred"
                        " while copying `{}' to `{}'".format(
                            src_dir,
                            dst_dir
                            )
                        )
                ret = 3

    return ret


def copy_file_or_directory(
        src_path,
        dst_dir,
        overwrite_files=False,
        clear_before_copy=False,
        dst_must_be_empty=True,
        verbose=True
        ):

    ret = 0

    if os.path.islink(src_path):
        ret = copy_symlink(
            src_path,
            dst_dir,
            # wayround_org.utils.path.join(dst_dir, os.path.basename(src_path)),
            overwrite_dst=overwrite_files,
            verbose=verbose
            )
    elif os.path.isfile(src_path):
        ret = copy_file(
            src_path,
            dst_dir,
            # wayround_org.utils.path.join(dst_dir, os.path.basename(src_path)),
            overwrite_dst=overwrite_files,
            verbose=verbose
            )
    else:
        if os.path.isdir(src_path):
            ret = copytree(
                src_path,
                dst_dir,
                overwrite_files=overwrite_files,
                clear_before_copy=clear_before_copy,
                dst_must_be_empty=dst_must_be_empty,
                verbose=verbose
                )

        else:
            if verbose:
                logging.error(
                    "Don't know what to do with: {}".format(src_path))
            ret = 1

    return ret


def isdirempty(dirname):
    return len(os.listdir(dirname)) == 0

is_dir_empty = isdirempty


def remove_if_exists(file_or_dir, remove_dead_lymlinks=False, verbose=True):

    ret = 0

    file_or_dir = wayround_org.utils.path.abspath(file_or_dir)

    if os.path.islink(file_or_dir):
        try:
            os.unlink(file_or_dir)
        except:
            if verbose:
                logging.exception(
                    "      can't remove link {}".format(file_or_dir)
                    )
            ret = 1

    if ret == 0 and os.path.exists(file_or_dir):

        if not os.path.islink(file_or_dir):

            if os.path.isdir(file_or_dir):
                try:
                    shutil.rmtree(file_or_dir)
                except:
                    if verbose:
                        logging.exception(
                            "Can't remove dir {}".format(file_or_dir)
                            )
                    ret = 1
            else:
                try:
                    os.unlink(file_or_dir)
                except:
                    if verbose:
                        logging.exception(
                            "      can't remove file {}".format(file_or_dir)
                            )
                    ret = 1

    return 0


def create_if_not_exists_dir(dirname, verbose=True):
    ret = 0
    if not os.path.exists(dirname):
        try:
            os.makedirs(dirname)
        except:
            if verbose:
                logging.error(
                    "Destination dir not exists and cant's be created")
            ret = 1
        else:
            ret = 0
    else:
        if os.path.isfile(dirname):
            if verbose:
                logging.error("Destination exists but is file")
            ret = 2
        elif os.path.islink(dirname):
            if verbose:
                logging.error("Destination exists but is link")
            ret = 3
        elif not os.path.isdir(dirname):
            if verbose:
                logging.error("Destination exists but is not dir")
            ret = 4
        else:
            ret = 0
    return ret


def cleanup_dir(dirname, verbose=True):

    ret = 0

    files = os.listdir(dirname)

    for i in range(len(files)):
        files[i] = wayround_org.utils.path.abspath(
            wayround_org.utils.path.join(dirname, files[i])
            )

    for i in files:
        if remove_if_exists(i, verbose=verbose) != 0:
            ret = 1

    return ret


def inderictory_copy_file(directory, file1, file2, verbose=True):

    file1 = os.path.basename(file1)
    file2 = os.path.basename(file2)

    f1 = os.path.join(directory, file1)
    f2 = os.path.join(directory, file2)

    if os.path.isfile(f1):
        if os.path.exists(f2):
            if verbose:
                logging.error("destination file or dir already exists")
        else:
            logging.info("copying {} to {}".format(f1, f2))
            try:
                shutil.copy(f1, f2)
            except:
                if verbose:
                    logging.exception("Error copying file")
    else:
        if verbose:
            logging.error("source file not exists")

    return


def files_recurcive_list(
        dirname,
        onerror=None,
        followlinks=False,
        exclude_paths=None,
        relative_to=None,
        mute=True,
        sort=False,
        acceptable_endings=None,
        print_found=False,
        list_symlincs=True,
        maxdepth=None,
        include_dirs=False
        ):

    s_sep = wayround_org.utils.path.select_s_sep(dirname)

    if relative_to and not isinstance(relative_to, str):
        raise ValueError("relative_to must be str or None")

    if maxdepth:

        if not isinstance(maxdepth, int):
            raise ValueError("maxdepth must be int or None")

        else:
            if maxdepth < 0:
                raise ValueError("maxdepth can't be lesser than zero")

    dirname = wayround_org.utils.path.normpath(dirname)

    absp = dirname.startswith(s_sep)
    dirname = wayround_org.utils.path.abspath(dirname)

    lst = []

    if not mute:
        logging.info(
            "Setting list of files in `{}' and it's subdirs".format(dirname)
            )

    if isinstance(exclude_paths, list):
        exclude_paths2 = list()

        for i in exclude_paths:

            f_path = None

            if i.startswith(s_sep):
                f_path = wayround_org.utils.path.abspath(i)
            else:
                f_path = wayround_org.utils.path.abspath(
                    wayround_org.utils.path.join(
                        dirname, i
                        )
                    )

            if ((dirname == os.path.sep and f_path.startswith(os.path.sep)) or
                    (f_path + os.path.sep).startswith(dirname + os.path.sep)):
                exclude_paths2.append(f_path)

        exclude_paths = exclude_paths2

    if not mute:
        logging.info("Walking...")

    dirname_path_length = wayround_org.utils.path.path_length(dirname)

    for dire, dirs, files in os.walk(
            dirname,
            onerror=onerror,
            followlinks=followlinks
            ):

        current_path_length = (
            wayround_org.utils.path.path_length(dire) - dirname_path_length
            ) + 1

        if maxdepth:
            if current_path_length == maxdepth:
                while len(dirs) != 0:
                    del(dirs[0])

        if sort:
            dirs.sort()
            files.sort()

        if isinstance(exclude_paths, list):

            for i in dirs[:]:

                f_path = wayround_org.utils.path.abspath(
                    wayround_org.utils.path.join(
                        dire, i
                        )
                    )

                if f_path in exclude_paths:
                    while i in dirs:
                        dirs.remove(i)

        for f in files:

            f_path = wayround_org.utils.path.join(dire, f)

            append = True

            if not list_symlincs and os.path.islink(f_path):
                append = False

            if append:
                if acceptable_endings:

                    found = False

                    for i in acceptable_endings:
                        if f.endswith(i):
                            found = True
                            break

                    if not found:
                        append = False

            if append:
                if print_found and not mute:
                    print("    {}".format(f_path))
                lst.append(f_path)

        if include_dirs:
            for f in dirs:

                f_path = wayround_org.utils.path.join(dire, f)

                append = True

                if not list_symlincs and os.path.islink(f_path):
                    append = False

                if append:
                    if print_found and not mute:
                        print("    {}".format(f_path))
                    lst.append(f_path)

        if not mute:
            pp = None
            if not relative_to:
                pp = dire
            else:
                pp = wayround_org.utils.path.relpath(dire, relative_to)

            wayround_org.utils.terminal.progress_write(
                "    ({} files): {}".format(len(lst), pp)
                )

    if not mute:
        wayround_org.utils.terminal.progress_write_finish()

    ret = lst

    if not relative_to:

        if not absp:

            p_path_s_l = len(dirname) + 1

            for i in range(len(ret)):
                ret[i] = ret[i][p_path_s_l:]

    else:

        for i in range(len(ret)):
            ret[i] = wayround_org.utils.path.relpath(ret[i], relative_to)

    return ret


def null_file(filename):
    ret = 0
    try:
        f = open(filename, 'w')
    except:
        ret = 1
    else:
        f.close()
    return ret


def get_dir_size(name):
    name = wayround_org.utils.path.abspath(name)
    if not os.path.isdir(name):
        raise OSError("Not a dir `{}'".format(name))

    size = 0

    lst = os.listdir(name)
    for i in ['.', '..']:
        if i in lst:
            lst.remove(i)

    lst.sort()

    for i in lst:
        f_pth = wayround_org.utils.path.abspath(name + os.path.sep + i)
        if not os.path.islink(f_pth):
            if os.path.isfile(f_pth):
                s = os.stat(f_pth)
                size += s.st_size
            elif os.path.isdir(f_pth):
                size += get_dir_size(f_pth)

    return size


def get_file_size(name):

    ret = None

    if not os.path.exists(name):
        ret = None
    else:
        if os.path.isfile(name):
            s = os.stat(name)
            ret = s.st_size
        elif os.path.isdir(name):
            ret = get_dir_size(name)

    return ret


def convert_symlink_to_file(path):

    ret = 0

    if os.path.islink(path):

        r_path = os.path.realpath(path)

        if os.path.isfile(r_path):
            os.unlink(path)
            shutil.copy2(r_path, path)
            ret = 0

        else:
            ret = 1

    return ret


def dereference_file(filename):

    ret = 0

    filename = wayround_org.utils.path.abspath(filename)

    if not os.path.exists(filename):
        ret = 1

    else:
        if os.path.isdir(filename):
            ret = 0
        else:
            if not os.path.islink(filename):
                ret = 0
            else:
                lnk = None
                dir_name = wayround_org.utils.path.abspath(
                    os.path.dirname(filename)
                    )

                try:
                    lnk = os.readlink(filename)
                except:
                    ret = 2
                else:
                    if lnk[0] != '/':
                        lnk = wayround_org.utils.path.abspath(
                            wayround_org.utils.path.join(dir_name, lnk)
                            )

                    os.unlink(filename)
                    shutil.copy2(lnk, filename)

    return ret


def dereference_files_in_dir(dirname):

    ret = 0

    dirname = wayround_org.utils.path.abspath(dirname)

    try:
        for dirpath, dirnames, filenames in os.walk(dirname):
            filenames.sort()
            dirnames.sort()
            dirpath = wayround_org.utils.path.abspath(dirpath)

            for i in filenames:
                if dereference_file(os.path.join(dirpath, i)) != 0:
                    logging.error(
                        "Could not dereference `{}'".format(
                            wayround_org.utils.path.relpath(
                                dirname,
                                os.getcwd()
                                )
                            )
                        )

                    ret = 1
    except:
        logging.exception("Dir files dereferencing exception")
        ret = 2

    return ret


def files_by_mask_copy_to_dir(in_dir, out_dir, mask='*.h'):

    in_dir = wayround_org.utils.path.abspath(in_dir)
    out_dir = wayround_org.utils.path.abspath(out_dir)

    ret = 0
    try:
        for dirpath, dirnames, filenames in os.walk(in_dir):
            filenames.sort()
            dirnames.sort()
            dirpath = wayround_org.utils.path.abspath(dirpath)

            for i in filenames:
                if fnmatch.fnmatch(i, mask):
                    shutil.copy2(
                        dirpath + os.path.sep + i,
                        out_dir + os.path.sep + i,
                        follow_symlinks=True
                        )

    except:
        logging.exception("Files by mask copy to dir exception")
        ret = 1

    return ret


class FDStatusWatcher:

    def __init__(self, on_status_changed=None):

        self._clear(init=True)

        self._on_status_changed = on_status_changed

    def __del__(self):
        self.stop()

    def _clear(self, init=False):

        if not init:
            if self.stat() != 'stopped':
                raise RuntimeError("Working. Cleaning restricted")

        if not init:
            if self._fd and self._poll:
                self._poll.unregister(self._fd)

        self._poll = None

        self._fd = None

        self._watching_thread = None

        self._stop_flag = False

        self.fd_status = None

        self._starting = False
        self._stopping = False

    def set_fd(self, fd):

        if not isinstance(fd, int):
            raise TypeError("`fd' must be int")

        logging.debug(
            "{}: Switching monitoring to new fd {}, previaus was {}".format(
                type(self).__name__,
                fd, self._fd
                )
            )

        worked = False
        if self.stat() == 'working':
            worked = True

            self.stop()

        self._fd = fd

        if worked:
            self.start()

    def get_fd(self):

        return self._fd

    def stop(self):

        if not self._starting and not self._stopping:

            self._stopping = True
            self._stop_flag = True
            self.wait('stopped')
            self._clear()
            self._stopping = False

    def start(self):

        threading.Thread(
            name="Thread Starting Socket Watcher",
            target=self._start
            ).start()

    def _start(self):

        if not isinstance(self._fd, int):
            raise TypeError("file descriptor must be given before start")

        if (not self._starting
                and not self._stopping
                and self.stat() == 'stopped'):

            self._starting = True

            self._poll = select.poll()
            self._poll.register(
                self._fd,
                select.POLLIN | select.POLLPRI | select.POLLOUT |
                select.POLLERR | select.POLLHUP | select.POLLNVAL
                )

            self._watching_thread = threading.Thread(
                name="Thread watching FD {}".format(self._fd),
                target=self._watching_method
                )

            self._watching_thread.start()

            self._starting = False

    def stat(self):

        ret = 'stopped'

        if bool(self._watching_thread):
            ret = 'working'
        else:
            ret = 'stopped'

        return ret

    def wait(self, what='stopped'):

        allowed_what = ['stopped', 'working']

        if not what in allowed_what:
            raise ValueError("`what' must be in {}".format(allowed_what))

        while True:
            time.sleep(0.1)
            if self.stat() == what:
                break

        return

    def _watching_method(self):

        while True:

            if self._stop_flag:
                break

            new_stat_lst = self._poll.poll(100)

            if len(new_stat_lst) > 0:

                new_stat_event = new_stat_lst[0][1]

                if new_stat_event != self.fd_status:
                    threading.Thread(
                        name="FD {} status changed to [{}]".format(
                            self._fd,
                            '|'.join(poll_stat_namer(new_stat_event))
                            ),
                        target=self._on_status_changed,
                        args=(self._fd, poll_stat_namer(new_stat_event),)
                        ).start()

                self.fd_status = new_stat_event

        self._watching_thread = None


def poll_stat_namer(event):

    if not isinstance(event, int):
        raise TypeError("`event' must be int, but it is `{}'".format(event))

    devided = poll_stat_devider(event)
    names = set()

    for i in POLL_CONSTS.keys():

        for j in devided:

            if POLL_CONSTS[i] == j:
                names.add(i)

    return list(names)


def poll_stat_devider(event):

    if not isinstance(event, int):
        raise TypeError("`event' must be int, but it is `{}'".format(event))

    devided = []

    for i in POLL_CONSTS:

        int_val = eval("select.{}".format(i))

        if int_val & event != 0:
            devided.append(int_val)

    return devided


def print_status_change(sock, stats):
    logging.info("File descriptor {} status changed to {}".format(sock, stats))


def which(name, under=None, exception_if_not_found=True):
    """
    define 'under' to exclude not acceptable PATH components
    """

    ret = None

    splitted_PATH = os.environ['PATH'].split(':')

    if isinstance(under, str):
        under = [under]

    if under is None:
        under = copy.copy(splitted_PATH)

    for i in range(len(under) - 1, -1, -1):
        ui = under[i]

        if not ui.endswith('/bin') and not ui.endswith('/sbin'):
            under.append(
                wayround_org.utils.path.join(
                    ui,
                    'bin'
                    )
                )
            under.append(
                wayround_org.utils.path.join(
                    ui,
                    'sbin'
                    )
                )

    for i in range(len(under) - 1, -1, -1):
        if not os.path.isdir(under[i]):
            del under[i]

    for i in under:

        n_f_n = wayround_org.utils.path.join(i, name)

        # print("checking: {}".format(n_f_n))

        if os.path.isfile(n_f_n):
            ret = n_f_n
            break

    if ret is None and exception_if_not_found:
        raise FileNotFoundError("`{}' executable not found".format(name))

    return ret


def checksumed_dir_redue(reducing_dir, reducable_dir, method='sha512'):

    reducing_dir_filelist = wayround_org.utils.file.files_recurcive_list(
        reducing_dir,
        onerror=None,
        followlinks=False,
        exclude_paths=None,
        relative_to=None,
        mute=False,
        sort=True,
        acceptable_endings=None,
        print_found=False,
        list_symlincs=False,
        maxdepth=None
        )

    reducing_dir_sums = wayround_org.utils.checksum.checksums_by_list(
        reducing_dir_filelist, method
        )

    reducable_dir_filelist = wayround_org.utils.file.files_recurcive_list(
        reducable_dir,
        onerror=None,
        followlinks=False,
        exclude_paths=None,
        relative_to=None,
        mute=False,
        sort=True,
        acceptable_endings=None,
        print_found=False,
        list_symlincs=False,
        maxdepth=None
        )

    reducable_dir_sums = wayround_org.utils.checksum.checksums_by_list(
        reducable_dir_filelist, method
        )

    for i in reducable_dir_filelist:

        for j in reducing_dir_filelist:

            if reducing_dir_sums[j] == reducable_dir_sums[i]:

                rel_path = \
                    wayround_org.utils.path.relpath(j, os.path.dirname(i))

                if os.path.isfile(i) or os.path.islink(i):
                    os.unlink(i)

                os.symlink(rel_path, i)

                break

    return 0





"""
Factory for quick constructing classes with attributes and checks

see also wayround_org.utils.lxml.checker_factory()
"""

import wayround_org.utils.types


def class_generate_attributes(class_obj, attributes):
    """
    Installs get_* and set_* functions to set or get _* values

    _* values are reset to None

    Generated methods can make use of format_* functions of same instance

    there are two cases for attributes:
    1. list of strings: names of attributes to create;
    2. list of 2-tuples:
        [0] - same as in first case;
        [1] - signal name to call on successful set operation.
    """

    case = None

    if wayround_org.utils.types.struct_check(
            attributes,
            {'t': list, '.': {'t': str}}
            ):
        case = 0

    if wayround_org.utils.types.struct_check(
            attributes,
            {'t': list, '.': {'t': tuple, '<': 2, '>': 2, '.': {'t': str}}}
            ):
        case = 1

    if case is None:
        raise ValueError(
            "`attributes' must be list of strings or list of tuple of strings"
            )

    for i in attributes:

        if isinstance(i, str):
            i = (i,)

        signal_emiter = ''

        if case == 1:
            signal_emiter = \
                "self.signal.emit('{signal_name}', self)".format(
                    signal_name=i[1]
                    )

        exec("""\
def set_{name}(self, value):
    \"""
    Assign value to self._{name} checking it with self.check_{name}() and
    formatting it with self.format_{name}() if it's exists
    \"""
    self.check_{name}(value)
    if hasattr(self, 'format_{name}') and callable(self.format_{name}):
        value = self.format_{name}(value)
    self._{name} = value

    {signal_emiter}

def get_{name}(self):
    \"""
    Retrieve value from self._{name} checking it with self.check_{name}()
    \"""
    ret = self._{name}
    self.check_{name}(ret)
    if hasattr(self, 'format_{name}') and callable(self.format_{name}):
        ret = self.format_{name}(ret)
    return ret

class_obj._{name} = None
class_obj.set_{name} = set_{name}
class_obj.get_{name} = get_{name}

del set_{name}
del get_{name}

""".format(name=i[0], signal_emiter=signal_emiter))

    return


def class_generate_check(class_obj, attributes):
    """
    Installs check function check(self, inst=None)

    New function checks each value in instance by getting them with get_*
    functions.

    New function can make use of logical_structure_check function to check
    instance sanity
    """

    if not wayround_org.utils.types.struct_check(
            attributes,
            {'t': list, '.': {'t': str}}
            ):
        raise ValueError("`attributes' must be list of strings")

    central_check = """\
def check(self, inst=None):
    if inst != None and type(self) != type(inst):
        raise ValueError("`inst' must be of same type as self")

    if inst == None:
        inst = self

"""

    for i in attributes:
        central_check += '    inst.get_{name}()\n'.format(name=i)

    central_check += """\
    if hasattr(self, 'logical_structure_check') and callable(self.logical_structure_check):
        inst.logical_structure_check()

"""

    exec("""\
{}

class_obj.check = check

del check
""".format(central_check))

    return


def class_generate_attributes_and_check(class_obj, attributes):
    class_generate_attributes(class_obj, attributes)
    class_generate_check(class_obj, attributes)
    return




import os
import copy
import grp
import pwd


def env_vars_edit(var_list, mode='copy'):
    """
    Environment preparations
    """

    ret = {}

    if mode == 'copy':
        ret = copy.deepcopy(os.environ)
    elif mode == 'clean':
        ret = {}
    else:
        raise ValueError("Invalid Value")

    if mode != 'clean':
        for i in list(var_list.keys()):
            if var_list[i] is None and i in ret:
                del(ret[i])
            else:
                ret[i] = var_list[i]

    return ret


def list_dirs_for_so_files():
    """
    Return list of system directories in which shared object (so) files are
    stored. basicly it is lines from /etc/ld.co.conf + splitted values from
    LD_LIBRARY_PATH environment variable
    """

    # TODO: todo

    ret = []

    return ret


def convert_gid_uid(gid, uid):

    if gid:

        if isinstance(gid, str):
            if gid.isnumeric():
                gid = int(gid)
            else:
                gid = grp.getgrnam(gid)[2]

    if uid:

        if isinstance(uid, str):
            if uid.isnumeric():
                uid = int(uid)
            else:
                uid = pwd.getpwnam(uid)[2]

    return gid, uid




import inspect


def render_help(object_dic, order_list=None, only_titles=False, key=None):

    # FIXME: complete doc
    """
    Render Python function __doc__ text to help

    object_dic - dict with <'name':callable> structure.
    order_list - order in whch outut must be created
    only_titles - extract only titles(first lines) from callables __doc__ members
    key -
    """

    key_list = []

    if isinstance(order_list, list):
        if set(list(object_dic.keys())) != set(order_list):
            raise ValueError(
                "object_dic keys must be equal to order_list\n\n{}\n\n{}".format(
                    repr(set(list(object_dic.keys()))),
                    repr(set(order_list))
                    )
                )

        key_list = order_list

    else:
        key_list = list(object_dic.keys())

    for i in key_list:
        if not isinstance(i, str):
            raise ValueError("all keys must be of type str")

    out = ''

    for i in key_list:

        txt = ''
        if not key:
            txt = inspect.getdoc(object_dic[i])
        else:
            txt = inspect.getdoc(key(object_dic[i]))

        if isinstance(txt, str):
            txt = txt.strip()

        if isinstance(txt, list):
            txt = '\n'.join(txt)
        elif txt is None:
            txt = "(No documentation)"
        elif not isinstance(txt, str):
            txt = str(txt)

        if only_titles:
            txt = str(txt)
            txt = '        ' + txt.splitlines()[0].strip()
        else:
            text = ''
            for j in txt.splitlines():
                text += "        {}\n".format(j)

            txt = text

        out += (
            """\
    {key}
{help}
""".format(
                key=i,
                help=txt
                )
            )

    return out




import logging
import os.path

import sqlalchemy.ext.declarative
import sqlalchemy.orm

import wayround_org.utils.path


class FileIndexer:

    Base = sqlalchemy.ext.declarative.declarative_base()

    class File(Base):

        __tablename__ = 'file'

        name = sqlalchemy.Column(
            sqlalchemy.UnicodeText,
            nullable=False,
            primary_key=True,
            default=''
            )

    def __init__(
            self,
            config_string,
            add_commit_items_no=200,
            del_commit_items_no=200,
            ):

        self._db_engine = \
            sqlalchemy.create_engine(
                config_string,
                echo=False
            )

        self.Base.metadata.bind = self._db_engine

        self.Base.metadata.create_all()

        self.add_commit_items_no = add_commit_items_no
        self.add_commit_counter = 0

        self.del_commit_items_no = del_commit_items_no
        self.del_commit_counter = 0

        try:
            self.session = sqlalchemy.orm.Session(bind=self._db_engine)
        except:
            self.session = None
            raise

        self.added_count = 0
        self.exists_count = 0
        self.deleted_count = 0

        return

    def __del__(self):

        self.close()

        return

    def close(self):
        if self.sess:
            self.session.commit()
            self.session.close()
            self.session = None
        return

    def commit(self):
        self.session.commit()
        return

    def is_name_exists(self, name):
        c = self.session.query(self.File).filter_by(name=name).count()
        return c != 0

    def delete(self, name):
        self.session.query(self.File).filter_by(name=name).delete()
        return

    def delete_missing(self, basedir):
        names = self.session.query(self.File).all()

        logging.info("Searching for missing files")

        for i in names:
            full_name = wayround_org.utils.path.abspath(
                wayround_org.utils.path.join(
                    basedir,
                    i.name
                    )
                )
            if not os.path.exists(full_name):

                logging.debug("`{}' not found -- deleting".format(full_name))

                self.session.delete(i)

                self.del_commit_counter += 1
                self.deleted_count += 1

                if self.del_commit_counter >= self.del_commit_items_no:
                    self.session.commit()
                    self.del_commit_counter = 0

    def add(self, name):

        if not self.is_name_exists(name):

            f = self.File()
            f.name = name

            self.session.add(f)

            self.add_commit_counter += 1
            self.added_count += 1

            if self.add_commit_counter >= self.add_commit_items_no:
                self.session.commit()
                self.add_commit_counter = 0

        else:
            self.exists_count += 1

        return

    def reset_counters(self):
        self.exists_count = 0
        self.added_count = 0

    def get_size(self):
        return self.session.query(self.File).count()

    def get_files(self):
        q = self.session.query(self.File).all()
        ret = []
        for i in q:
            ret.append(i.name)
        return ret




import copy
import inspect
import logging
import sys

import wayround_org.utils.error
import wayround_org.utils.getopt
import wayround_org.utils.logging


NO_DOCUMENTATION = '(No documentation)'

DO_NOT_PRINT_EXIT_MESSAGE = 'showing help cmd list, no print'


def logging_setup(loglevel='INFO'):

    loglevel = loglevel.upper()

    # Logging settings
    for i in [
            (logging.CRITICAL, '[c]'),
            (logging.ERROR, '[e]'),
            (logging.WARN, '[w]'),
            (logging.WARNING, '[w]'),
            (logging.INFO, '[i]'),
            (logging.DEBUG, '[d]')
            ]:
        logging.addLevelName(i[0], i[1])
    del i

    opts = wayround_org.utils.getopt.getopt_keyed(sys.argv[1:])[0]

    # Setup logging level and format
    log_level = loglevel

    if '--loglevel' in opts:
        log_level_u = opts['--loglevel'].upper()

        if not log_level_u in wayround_org.utils.logging.LEVEL_NAMES:
            print("[e] Wrong --loglevel parameter")
        else:
            log_level = log_level_u

        del(opts['--loglevel'])
        del(log_level_u)

    logging.basicConfig(
        format="%(levelname)s %(message)s",
        level=log_level
        )

    return


def command_processor(
        command_name,
        commands,
        opts_and_args_list,
        additional_data
        ):

    opts, args = wayround_org.utils.getopt.getopt_keyed(opts_and_args_list)

    ret = dict(
        code=0,
        message='Default Exit',
        main_message="No Error"
        )

    args_l = len(args)

    subtree = commands

    level_depth = []

    for i in range(args_l):

        ii = args[i]

        if not ii in subtree:
            ret = dict(
                code=100,
                message='error',
                main_message="invalid command or subsection name"
                )
            break

        subtree = subtree[ii]
        level_depth.append(ii)

        if callable(subtree):
            break

    if ret['code'] != 0:
        pass
    else:

        args = args[len(level_depth):]

        args_l = len(args)

        show_help = '--help' in opts
        show_list = '--help-list' in opts

        if callable(subtree):

            if show_help:

                ret = {
                    'code': 0,
                    'message': "showing help",
                    'main_message': _format_command_help(level_depth, subtree)
                    }

            elif show_list:

                # TODO: print error

                ret = {
                    'code': 1,
                    'message': "error",
                    'main_message':
                        "--help-list param not applicable to final command"
                    }

            else:

                try:
                    res = subtree(
                        level_depth,
                        opts,
                        args,
                        additional_data
                        )
                except BrokenPipeError:
                    ret = dict(
                        code=1,
                        message="BrokenPipeError"
                        )
                except KeyboardInterrupt:
                    ret = dict(
                        code=1,
                        message='error',
                        main_message="Interrupted With Keyboard"
                        )
                except:
                    e = sys.exc_info()

                    ex_txt = wayround_org.utils.error.return_exception_info(
                        e,
                        tb=True
                        )

                    ret = dict(
                        code=1,
                        message='error',
                        main_message=(
                            "Error while executing command: {}\n{}".format(
                                ' '.join(level_depth),
                                ex_txt
                                )
                            )
                        )

                else:

                    if isinstance(res, int):
                        txt = None

                        if res == 0:
                            txt = 'No errors'
                        else:
                            txt = 'Some error'

                        ret = dict(
                            code=res,
                            message=txt
                            )
                    elif isinstance(res, dict):

                        ret = dict(
                            code=res['code'],
                            message='error',
                            main_message=res['message']
                            )

                    else:
                        ret = dict(
                            code=1,
                            message='error',
                            main_message=(
                                "Command returned not integer and not dict\n"
                                "(forcing exit code 1 and 'error' message)\n"
                                "Actually it has returned(type:{}):\n{}".format(
                                    type(res),
                                    res
                                    )
                                )
                            )

        elif isinstance(subtree, dict):

            if show_help:

                ret = {
                    'code': 0,
                    'message': "showing help",
                    'main_message': _format_command_level_help(
                        subtree,
                        level_depth
                        )
                    }

            elif show_list:

                ret = {
                    'code': 1,
                    'message': DO_NOT_PRINT_EXIT_MESSAGE,
                    'main_message': _format_command_list(
                        subtree,
                        level_depth
                        )
                    }

            else:

                ret = dict(
                    code=1,
                    message='error',
                    main_message=(
                        "Callable command not supplied. "
                        "May be try use --help param."
                        )
                    )

        else:
            raise ValueError("invalid command tree")

    return ret


def program(command_name, commands, additional_data=None):
    """
    command_name used only for help rendering purposes, so if not given --
    program name will not be rendered in help.

    this function uses command_processor() for command_processing, see it's
    documentation for explanations on parameters
    """

    ret = command_processor(
        command_name,
        commands,
        sys.argv[1:],
        additional_data
        )

    if ret['message'] != DO_NOT_PRINT_EXIT_MESSAGE:

        mm = ret.get('main_message')

        if mm is not None:
            logging.info('{}'.format(mm))

        logging.info(
            "Exit Code: {} ({})".format(ret['code'], ret['message'])
            )

    return ret['code']


def _format_command_help(level_depth, function):

    command_text = inspect.getdoc(function)

    if not isinstance(command_text, str):
        command_text = NO_DOCUMENTATION

    command_name_text = ' '.join(level_depth)

    ret = """\
Usage: {command_name_text} [options] [parameters]

{command_text}

""".format(
        command_text=command_text,
        command_name_text=command_name_text
        )

    return ret


def _format_command_list(subtree, level_depth):
    ret = None
    level = subtree

    if not hasattr(level, 'keys'):
        level = None

    if level is None:
        ret = None
    else:
        print('\n'.join(level.keys()))
        ret = ''

    return


def _format_command_level_help(subtree, level_depth):

    this_tree_help = NO_DOCUMENTATION
    command_name_text = ' '.join(level_depth)
    sections_text = ''
    subcommands_text = ''
    commands_text = ''

    if '_help' in subtree:
        this_tree_help = subtree['_help']

    for i in subtree.keys():

        if i == '_help':
            continue

        if callable(subtree[i]) or not isinstance(subtree[i], dict):
            continue

        command_help_text = NO_DOCUMENTATION

        if '_help' in subtree[i]:
            command_help_text = subtree[i]['_help']

        if isinstance(command_help_text, str):
            command_help_text = command_help_text.splitlines()[0].strip()

        sections_text += """\
    {cmd_name}
        {cmd_short_descr}

""".format(
            cmd_name=i,
            cmd_short_descr=command_help_text)

    for i in subtree.keys():

        if i == '_help':
            continue

        if not callable(subtree[i]):
            continue

        command_help_text = inspect.getdoc(subtree[i])

        if isinstance(command_help_text, str):
            command_help_text = command_help_text.splitlines()[0].strip()

        if not isinstance(command_help_text, str):
            command_help_text = NO_DOCUMENTATION

        commands_text += """\
    {cmd_name}
        {cmd_short_descr}

""".format(
            cmd_name=i,
            cmd_short_descr=command_help_text)

    this_tree_help_text = ''
    if this_tree_help != NO_DOCUMENTATION:
        this_tree_help_text = """\
{this_tree_help}
""".format(this_tree_help=this_tree_help)

    sub_sect_help_text = ''
    if sections_text != '':
        sub_sect_help_text = """\
sections:
{sect_text}
""".format(sect_text=sections_text)

    sub_comm_help_text = ''
    if commands_text != '':
        sub_comm_help_text = """\
commands:
{cmds_text}
""".format(cmds_text=commands_text)

    ret = "Usage: {command_name_text} [options] [parameters]\n".format(
        command_name_text=command_name_text
        )

    if this_tree_help_text != '':
        ret += '\n{}'.format(this_tree_help_text)

    if sub_sect_help_text != '':
        ret += '\n{}'.format(sub_sect_help_text)

    if sub_comm_help_text != '':
        ret += '\n{}'.format(sub_comm_help_text)

    ret += """\
    --help         help for section or command
    --help-list    list commands in section
"""

    print(ret)

    return




"""
Module for organizing persistent supporting flags for files

This can be used to create any sort of flags

YAML is used for storing data in flags

flags can be also used as simple raw files
"""

import copy
import os.path

import yaml

import wayround_org.utils.path
import wayround_org.utils.threading
import wayround_org.utils.types_presets


def verify_flag_name(flagged_file, flag_name):
    """
    function for checking whatever user supplied flag name supported by
    FlaggedFile class instance
    """
    if not isinstance(flagged_file, FlaggedFile):
        raise TypeError("`flagged_object' must be inst of FlaggedFile")

    if not isinstance(flag_name, str):
        raise TypeError("flag name must be str")

    if flag_name not in flagged_file.possible_flags:
        raise ValueError("invalid flag name")
    return


def check_formatted_access_to_raw_flag(flagged_file, flag_name):

    verify_flag_name(flagged_file, flag_name)

    if flag_name in flagged_file.raw_flags:
        raise ValueError(
            "trying to use formatted access to raw flag `{}'".format(
                flag_name
                )
            )
    return


class FlaggedFile:
    """
    Use this class for creating flags
    """

    def __init__(
            self,
            path,
            basename,
            possible_flags,
            raw_flags=None,
            object_locker=None
            ):
        """
        path - directory in which to store flags
        basename - base name to which flag extensions will be appended
        possible_flags - list (or set) of flags which supported by  instance of
            this class
        raw_flags - list (or set) of flags, internal structure of which should
            not be considered in YAML markup. so trying to use on them YAML
            methods of this class will raise exception.
        object_locker - instance of ObjectLocker to use for sefe access to
            flag files. use None for automatic ceation of this
        """

        self._path = None
        self._basename = None

        if raw_flags is None:
            raw_flags = set()

        if not isinstance(possible_flags, (list, set)):
            raise TypeError("`possible_flags' must be list or set")

        if not isinstance(raw_flags, (list, set)):
            raise TypeError("`raw_flags' must be list, set or None")

        self.path = path
        self.basename = basename
        self.possible_flags = set(possible_flags)
        self.raw_flags = set(raw_flags)

        for i in self.raw_flags:
            if i not in self.possible_flags:
                print(
                    "warning (FlaggedFile):"
                    " {} not in possible_flags".format(i)
                    )

        self.possible_flags |= self.raw_flags

        if object_locker is None:
            object_locker = wayround_org.utils.threading.ObjectLocker()

        self.object_locker = object_locker

        for i in possible_flags:
            setattr(self, '{}_path'.format(i), self.gen_flag_path(i))

        return

    def get_flags_list(self):
        return copy.copy(self.possible_flags)

    @property
    def path(self):
        """
        returns directory path with which this instance is working now.
        this is wrapped in property to ensure correct value is passed to it
        """
        return self._path

    @path.setter
    def path(self, value):
        """
        change path parameter
        """
        if not isinstance(value, str):
            raise TypeError("`path' must be str")
        self._path = value
        return

    @property
    def basename(self):
        return self._basename

    @basename.setter
    def basename(self, value):
        if not isinstance(value, str):
            raise TypeError("`basename' must be str")
        self._basename = value
        return

    def unset_all_flags(self):
        for i in self.possible_flags:
            self.unset_flag(i)
        return

    def gen_flag_path(self, name):
        verify_flag_name(self, name)
        ret = wayround_org.utils.path.join(
            self.path,
            '{}.{}'.format(self.basename, name)
            )
        return ret

    def get_flag_path(self, name):
        verify_flag_name(self, name)
        ret = getattr(self, '{}_path'.format(name))
        return ret

    def get_flag_lock(self, name):
        ret = self.object_locker.get_lock(self.get_flag_path(name))
        return ret

    def get_is_flag_locked(self, name):
        ret = self.object_locker.get_is_locked(self.get_flag_path(name))
        return ret

    def get_is_flag_set(self, name):
        ret = os.path.isfile(self.get_flag_path(name))
        return ret

    def open_flag(self, name, flags='r'):
        ret = open(self.get_flag_path(name), flags)
        return ret

    def write_flag_from_persistent_variable(
            self,
            name,
            pv,
            stop_event=None,
            bs=2 * 1024**2
            ):
        """
        ret: True - success, None - stopped
        """

        if not isinstance(
                pv,
                wayround_org.utils.pm.PersistentVariable
                ):
            raise TypeError(
                "must be inst of "
                "wayround_org.utils.pm.PersistentVariable"
                )

        ret = None

        with self.get_flag_lock(name):
            with self.open_flag(name, 'bw') as f1:
                with pv.open('br') as f2:
                    while True:
                        if stop_event is not None and stop_event.is_set():
                            ret = None
                            break
                        b = f2.read(bs)
                        if len(b) == 0:
                            ret = True
                            break
                        f1.write(b)
        return ret

    def set_flag(self, name):
        verify_flag_name(self, name)
        if name in self.raw_flags:
            f_path = self.get_flag_path(name)
            if not self.get_is_flag_set(name):
                with self.get_flag_lock(name):
                    with self.open_flag(name, 'wb'):
                        pass
        else:
            data = self.get_flag_data(name)
            self.set_flag_data(name, data)
        return

    def set_flag_data(self, name, data):

        check_formatted_access_to_raw_flag(self, name)

        with self.get_flag_lock(name):
            with self.open_flag(name, 'w') as f:
                f.write(yaml.dump(data))

        return

    def get_flag_data(self, name):

        check_formatted_access_to_raw_flag(self, name)

        ret = None

        if self.get_is_flag_set(name):
            with self.get_flag_lock(name):
                with self.open_flag(name, 'r') as f:
                    ret = yaml.load(f.read())

        return ret

    def get_flag_size(self, name):
        verify_flag_name(self, name)
        ret = os.stat(self.get_flag_path(name)).st_size
        return ret

    def unset_flag(self, name):
        with self.get_flag_lock(name):
            if self.get_is_flag_set(name):
                os.unlink(self.get_flag_path(name))
        return

    def get_bool(self, name):
        ret = self.get_flag_data(name) is True
        return ret

    def set_bool(self, name, value=True):
        if not isinstance(value, bool):
            raise TypeError("`{}' value must be bool".format(name))
        self.set_flag_data(name, value)
        return

    def get_str(self, name):
        ret = self.get_flag_data(name)
        if not isinstance(ret, str):
            ret = None
        return ret

    def set_str(self, name, value):
        if not isinstance(value, str):
            raise TypeError("`{}' value must be str".format(name))
        self.set_flag_data(name, value)
        return

    def set_str_n(self, name, value):
        """
        Same as set_str, but allows None as value
        """
        if value is not None and not isinstance(value, str):
            raise TypeError("`{}' value must be str or None".format(name))
        self.set_flag_data(name, value)
        return

    def get_int(self, name):
        ret = self.get_flag_data(name)
        if not isinstance(ret, int):
            ret = None
        return ret

    def set_int(self, name, value):
        if not isinstance(value, int):
            raise TypeError("`{}' value must be int".format(name))
        self.set_flag_data(name, value)
        return

    def set_int_n(self, name, value):
        """
        Same as set_int, but allows None as value
        """
        if value is not None and not isinstance(value, int):
            raise TypeError("`{}' value must be int or None".format(name))
        self.set_flag_data(name, value)
        return

    def get_str_list(self, name):
        ret = self.get_flag_data(name)
        if not wayround_org.utils.types_presets.is_list_of_str(ret):
            ret = []
        return ret

    def set_str_list(self, name, value):
        if not wayround_org.utils.types_presets.is_list_of_str(value):
            raise TypeError("`{}' value must be list of str".format(name))
        self.set_flag_data(name, value)
        return

    def add_str_list(self, name, value):
        self.set_str_list(name, self.get_str_list(name) + [value])
        return

    def get_str_set(self, name):
        ret = self.get_flag_data(name)
        if not wayround_org.utils.types_presets.is_set_of_str(ret):
            ret = set()
        return ret

    def set_str_set(self, name, value):
        if not wayround_org.utils.types_presets.is_set_of_str(value):
            raise TypeError("`{}' value must be set of str".format(name))
        self.set_flag_data(name, value)
        return

    def add_str_set(self, name, value):
        self.set_str_list(
            name,
            self.get_str_list(name) + [value]
            )
        return

    def get_int_list(self, name):
        ret = self.get_flag_data(name)
        if not wayround_org.utils.types_presets.is_list_of_int(ret):
            ret = []
        return ret

    def set_int_list(self, name, value):
        if not wayround_org.utils.types_presets.is_list_of_int(value):
            raise TypeError("`{}' value must be list of int".format(name))
        self.set_flag_data(name, value)
        return

    def set_int_list_n(self, name, value):
        if (value is not None
            and not wayround_org.utils.types_presets.is_list_of_int(value)
            ):
            raise TypeError(
                "`{}' value must be None or list of int".format(name)
                )
        self.set_flag_data(name, value)
        return




import copy
import fnmatch
import logging
import re
import os.path


def remove_all_values(lst, lst2):
    """
    Removes all values of list lst2 from list lst
    """
    for i in lst2:
        while i in lst:
            lst.remove(i)
    return


def list_lstrip(lst, lst2):
    lst = copy.copy(lst)
    while (len(lst) > 0
           and lst[0] in lst2):
        del lst[0]
    return lst


def list_rstrip(lst, lst2):
    lst = copy.copy(lst)
    while (len(lst) > 0
           and lst[-1] in lst2):
        del lst[-1]
    return lst


def list_strip(lst, lst2):
    return list_lstrip(list_rstrip(copy.copy(lst), lst2), lst2)


def list_lower(lst):
    lst2 = []
    for i in lst:
        lst2.append(i.lower())
    return lst2


def list_upper(lst):
    lst2 = []
    for i in lst:
        lst2.append(i.upper())
    return lst2


def list_sort(lst, cmp=None):

    lst_l = len(lst)

    i = -1
    j = -1
    x = None

    if lst_l > 1:
        while True:
            if i == lst_l - 2:
                break
            i += 1
            j = i
            while True:

                if j == lst_l - 1:
                    break
                j += 1

                if cmp is None:
                    if lst[i] > lst[j]:
                        x = lst[i]
                        lst[i] = lst[j]
                        lst[j] = x
                else:
                    cmp_r = cmp(lst[i], lst[j])
                    if cmp_r == +1:
                        x = lst[i]
                        lst[i] = lst[j]
                        lst[j] = x

    return


def list_remove_empty_lines(lst):
    ret = []
    for i in lst:
        if i != '':
            ret.append(i)
    return ret


def list_remove_duplicated_lines(lst):
    _s = set()
    _l = list()
    for i in lst:
        if not i in _s:
            _s.add(i)
            _l.append(i)
    ret = _l
    return ret


def list_strip_lines(lst):
    ret = []
    for i in lst:
        ret.append(i.strip())
    return ret


def list_strip_filelines(lst):
    ret = []
    for i in lst:
        ret.append(i.strip('\n'))
    return ret


def list_strip_remove_empty_remove_duplicated_lines(lst):
    """
    Do some actions with list of lines

    Do not use this function for file lists,
    use filelist_strip_remove_empty_remove_duplicated_lines
    """
    ret = list_strip_lines(lst)
    ret = list_remove_empty_lines(ret)
    ret = list_remove_duplicated_lines(ret)
    return ret


def filelist_strip_remove_empty_remove_duplicated_lines(lst):
    """
    Does not strips spaces from file names.

    Use this function for file lists,
    not list_strip_remove_empty_remove_duplicated_lines
    """
    ret = list_strip_filelines(lst)
    ret = list_remove_empty_lines(ret)
    ret = list_remove_duplicated_lines(ret)
    return ret


def list_filter_list_white_or_black(
        in_str_list, in_filters_lst, white=True
        ):

    ret = []

    for i in in_str_list:
        i = str(i)

        for j in in_filters_lst:

            if white:
                if re.match(j, i):
                    ret.append(i)
                    break
            else:
                if not re.match(j, i):
                    ret.append(i)
                    break

    return ret


def filter_text_parse(filter_text, show_errors=False):
    """
    Returns list of command structures

    ret = [
        dict(
            action   = '-' or '+',
            function = <depends on subject> (no spaces allowed),
            data     = <depends on subject> (can contain spaces)
            )
        ]

    """
    ret = []

    lines = filter_text.splitlines()

    for i in lines:
        if i != '' and not i.isspace() and not i.startswith('#'):
            struct = i.split(' ', maxsplit=2)
            if not len(struct) == 3:
                if show_errors:
                    logging.error("Wrong filter line: `{}'".format(i))
                ret = None
                break
            else:
                struct = dict(
                    action=struct[0],
                    function=struct[1],
                    data=struct[2]
                    )
                ret.append(struct)

    return ret


def filter_list(input_list, filter_text):
    """
    Filters supplied list with supplied filter

    subjects not in check_for_subjects will always be positive (but can be
    filtered out by proper leading rules)
    """

    ret = []

    inp_list = set(copy.copy(input_list))
    out_list = copy.copy(inp_list)

    filters = filter_text_parse(filter_text)

    for f in filters:

        action = f['action']
        function = f['function']
        no = False
        cs = True
        data = f['data']

        if not action in ['+', '-']:
            logging.error("Wrong action: `{}'".format(action))
            ret = 10
            break

        if function.startswith('!'):
            no = True
            function = function[1:]

        if function.endswith('!'):
            cs = False
            function = function[:-1]

        if not function in ['begins', 'contains', 'ends', 'fm', 'bfm', 're']:
            logging.error(
                "Wrong function : `{}'".format(function)
                )
            ret = 3
            break

        if not isinstance(ret, int):

            working_list = None

            if action == '+':
                working_list = copy.copy(inp_list)

            elif action == '-':
                working_list = copy.copy(out_list)
            else:
                raise Exception("Programming Error")

            for item in working_list:

                working_item = item

                if not cs:
                    working_item = working_item.lower()

                matched = False

                if function == 'begins':
                    working_data = data
                    if not cs:
                        working_data = working_data.lower()
                    matched = working_item.startswith(working_data)

                elif function == 'contains':
                    working_data = data
                    if not cs:
                        working_data = working_data.lower()
                    matched = working_item.find(working_data) != -1

                elif function == 'end':
                    working_data = data
                    if not cs:
                        working_data = working_data.lower()
                    matched = working_item.endswith(working_data)

                elif function == 're':
                    working_data = data
                    flags = 0
                    if not cs:
                        flags |= re.IGNORECASE
                    matched = \
                        re.match(working_data, working_item, flags) is not None

                elif function == 'fm':
                    working_data = data
                    if not cs:
                        working_data = working_data.lower()
                    matched = fnmatch.fnmatch(working_item, working_data)

                elif function == 'bfm':
                    working_data = data
                    if not cs:
                        working_data = working_data.lower()
                    matched = fnmatch.fnmatch(
                        os.path.basename(working_item),
                        working_data
                        )

                else:
                    raise Exception("Programming error")

                if no:
                    matched = not matched

                if matched:

                    if action == '+':
                        out_list.add(item)

                    elif action == '-':
                        if item in out_list:
                            out_list.remove(item)

                    else:
                        raise Exception("Programming error")

                else:
                    pass

    if not isinstance(ret, int):
        ret = out_list

    if isinstance(ret, set):
        ret = list(ret)

    return ret




import logging


def getopt(args):
    """
    Parser as command line arguments and options

    It's not compatible with CPython getopt module. It's work differently in
    many ways.

    Parameter must be list of strings: Simple sys.argv will go.

    Example:

       getopt.getopt(['a', 'b', 'c', 'd=123', 'dd=123', '-a',
                       '3', '-b=3', '--c=4', '--long=5',
                       '---strange=6', '--', '-e=7'])

       ([('-a', None),
         ('-b', '3'),
         ('--c', '4'),
         ('--long', '5'),
         ('---strange', '6')],
        ['a', 'b', 'c', 'd=123', 'dd=123', '3', '-e=7'])

    """

    # TODO: add '--' delimiter support

    ret_args = []
    ret_opts = []

    if not isinstance(args, list):
        raise TypeError

    all_args = False

    len_args = len(args)

    i = 0

    while True:

        if i == len_args:
            break

        if all_args:

            ret_args.append(args[i])

        else:

            args_i_len = len(args[i])

            if args_i_len > 1:

                if args[i] == '--':
                    all_args = True
                else:

                    if args[i].startswith('-'):

                        eq_pos = args[i].find('=')
                        if eq_pos != -1:
                            ret_opts.append(
                                (args[i][:eq_pos], args[i][eq_pos + 1:])
                                )
                        else:
                            ret_opts.append((args[i], None))

                    else:
                        ret_args.append(args[i])

            else:
                ret_args.append(args[i])

        i += 1

    return ret_opts, ret_args


def getopt_keyed(args):
    opts, args = getopt(args)

    opts_k = {}
    for i in opts:
        opts_k[i[0]] = i[1]

    return opts_k, args


def _opt_strip(opt):
    return opt.lstrip('!').rstrip('=')


def _opts_strip(opts_list):

    ret = []

    for i in opts_list:
        ret.append(_opt_strip(i))

    return ret


def check_options(opts, opts_list, mute=False):
    """
    opts - list of opts returned by getopt (i.e.)
    opts_list - list of acceptable options:
          starting with '!' - required option
          ending with '=' - must have parameter
    """

    ret = 0

    for i in opts:
        if not i in _opts_strip(opts_list):

            if not mute:
                logging.error("option not supported: {}".format(i))

            ret += 1

    for i in opts_list:

        i_stripped = _opt_strip(i)

        required = i.startswith('!')
        value_required = i.endswith('=')

        if required and not i_stripped in opts:

            if not mute:
                logging.error(
                    "required parameter absent: {}".format(i_stripped)
                    )

            ret += 1

        if i_stripped in opts:
            if value_required and opts[i_stripped] is None:

                if not mute:
                    logging.error(
                        "parameter `{}' must have value".format(i_stripped)
                        )

                ret += 1

    return ret




import regex

ALPHA_RE = r'(\p{L})'

DIGIT_RE = r'[0-9]'

HEXDIG_RE = r'[0-9a-fA-F]'

'''
   pct-encoded   = "%" HEXDIG HEXDIG
'''
PCT_ENCODED_RE = r'(%{hexdig}{hexdig})'.format(hexdig=HEXDIG_RE)

'''
   unreserved    = ALPHA / DIGIT / "-" / "." / "_" / "~"
'''
UNRESERVED_RE = r'({alpha}|{digit}|-|\.|_|~)'.format(
    alpha=ALPHA_RE,
    digit=DIGIT_RE
    )

'''
   gen-delims    = ":" / "/" / "?" / "#" / "[" / "]" / "@"
'''
GEN_DELIMS_RE = r'[\:/\?\#\[\]@]'
'''
   sub-delims    = "!" / "$" / "&" / "'" / "(" / ")"
                 / "*" / "+" / "," / ";" / "="
'''
SUB_DELIMS_RE = r"[\!$&'()*+,;=]"

'''
   reserved      = gen-delims / sub-delims
'''
RESERVED_RE = r'({gen_delims}|{sub_delims})'.format(
    gen_delims=GEN_DELIMS_RE,
    sub_delims=SUB_DELIMS_RE
    )

'''
   pchar         = unreserved / pct-encoded / sub-delims / ":" / "@"
'''
PCHAR_RE = r'({unreserved}|{pct_encoded}|{sub_delims}|\:\@)'.format(
    unreserved=UNRESERVED_RE,
    pct_encoded=PCT_ENCODED_RE,
    sub_delims=SUB_DELIMS_RE
    )

'''
   query         = *( pchar / "/" / "?" )
'''
QUERY_RE = r'(({pchar}|\/|\?)*)'.format(pchar=PCHAR_RE)

'''
   fragment      = *( pchar / "/" / "?" )
'''
FRAGMENT_RE = r'(({pchar}|\/|\?)*)'.format(pchar=PCHAR_RE)

'''
   segment       = *pchar
'''
SEGMENT_RE = r'({pchar})*'.format(pchar=PCHAR_RE)

'''
   segment-nz    = 1*pchar
'''
SEGMENT_NZ_RE = r'(({pchar}){{1}})'.format(pchar=PCHAR_RE)

'''
   segment-nz-nc = 1*( unreserved / pct-encoded / sub-delims / "@" )
                 ; non-zero-length segment without any colon ":"
'''
SEGMENT_NZ_NC_RE = r'(({unreserved}|{pct_encoded}|{sub_delims}|@){{1}})'.format(
    unreserved=UNRESERVED_RE,
    pct_encoded=PCT_ENCODED_RE,
    sub_delims=SUB_DELIMS_RE
    )

'''
   path-abempty  = *( "/" segment )
'''
PATH_ABEMPTY_RE = r'(\/{segment})*'.format(segment=SEGMENT_RE)

'''
   path-absolute = "/" [ segment-nz *( "/" segment ) ]
'''
PATH_ABSOLUTE_RE = r'(\/({segment_nz}(\/{segment})*)?)'.format(
    segment_nz=SEGMENT_NZ_RE,
    segment=SEGMENT_RE
    )

'''
   path-noscheme = segment-nz-nc *( "/" segment )
'''
PATH_NOSCHEME_RE = r'({segment_nz_nc}(\/{segment})*)'.format(
    segment_nz_nc=SEGMENT_NZ_NC_RE,
    segment=SEGMENT_RE
    )

'''
   path-rootless = segment-nz *( "/" segment )
'''
PATH_ROOTLESS_RE = r'({segment_nz}(\/{segment})*)'.format(
    segment_nz=SEGMENT_NZ_RE,
    segment=SEGMENT_RE
    )

'''
   path-empty    = 0<pchar>
'''
PATH_EMPTY_RE = r'{pchar}{{0}}'.format(pchar=PCHAR_RE)

'''
   path          = path-abempty    ; begins with "/" or is empty
                 / path-absolute   ; begins with "/" but not "//"
                 / path-noscheme   ; begins with a non-colon segment
                 / path-rootless   ; begins with a segment
                 / path-empty      ; zero characters
'''
PATH_RE = (r'({path_abempty}|{path_absolute}|'
           r'{path_noscheme}|{path_rootless}|{path_empty})').format(
    path_abempty=PATH_ABEMPTY_RE,
    path_absolute=PATH_ABSOLUTE_RE,
    path_noscheme=PATH_NOSCHEME_RE,
    path_rootless=PATH_ROOTLESS_RE,
    path_empty=PATH_EMPTY_RE
    )

'''
   reg-name      = *( unreserved / pct-encoded / sub-delims )
'''
REG_NAME_RE = r'(({unreserved}|{pct_encoded}|{sub_delims})*)'.format(
    unreserved=UNRESERVED_RE,
    pct_encoded=PCT_ENCODED_RE,
    sub_delims=SUB_DELIMS_RE
    )

'''
   dec-octet     = DIGIT                 ; 0-9
                 / %x31-39 DIGIT         ; 10-99
                 / "1" 2DIGIT            ; 100-199
                 / "2" %x30-34 DIGIT     ; 200-249
                 / "25" %x30-35          ; 250-255
'''
DEC_OCTET_RE = (r'({digit}|([1-9]{digit})'
                r'|(1{digit}{{2}})|(2[0-4]{digit})|(25[05]))').format(
    digit=DIGIT_RE
    )

'''
   IPv4address   = dec-octet "." dec-octet "." dec-octet "." dec-octet
'''
IPV4ADDRESS_RE = r'({dec_octet}\.{dec_octet}\.{dec_octet}\.{dec_octet})'.format(
    dec_octet=DEC_OCTET_RE
    )

'''
   h16           = 1*4HEXDIG
'''
H16_RE = r'{hexdig}{{1:4}}'.format(hexdig=HEXDIG_RE)

'''
   ls32          = ( h16 ":" h16 ) / IPv4address
'''
LS32_RE = r'(({h16}\:{h16})|{ipv4address})'.format(
    h16=H16_RE,
    ipv4address=IPV4ADDRESS_RE
    )

'''
   IPv6address   =                            6( h16 ":" ) ls32
                 /                       "::" 5( h16 ":" ) ls32
                 / [               h16 ] "::" 4( h16 ":" ) ls32
                 / [ *1( h16 ":" ) h16 ] "::" 3( h16 ":" ) ls32
                 / [ *2( h16 ":" ) h16 ] "::" 2( h16 ":" ) ls32
                 / [ *3( h16 ":" ) h16 ] "::"    h16 ":"   ls32
                 / [ *4( h16 ":" ) h16 ] "::"              ls32
                 / [ *5( h16 ":" ) h16 ] "::"              h16
                 / [ *6( h16 ":" ) h16 ] "::"
'''
IPV6ADDRESS_RE = (
    r'('
    r'({h16}\:){{6}}{ls32}'
    r'\:\:({h16}\:){{5}}{ls32}'
    r'({h16})?\:\:({h16}\:){{4}}{ls32}'
    r'(({h16}\:){{1}}{h16})?\:\:({h16}\:){{3}}{ls32}'
    r'(({h16}\:){{2}}{h16})?\:\:({h16}\:){{2}}{ls32}'
    r'(({h16}\:){{3}}{h16})?\:\:({h16}\:){{1}}{ls32}'
    r'(({h16}\:){{4}}{h16})?\:\:({h16}\:){{0}}{ls32}'
    r'(({h16}\:){{5}}{h16})?\:\:({h16}\:){{0}}{ls32}'
    r'(({h16}\:){{6}}{h16})?\:\:({h16}\:){{0}}{ls32}'
    r')'
    ).format(
        h16=H16_RE,
        ls32=LS32_RE
        )

'''
   IPvFuture     = "v" 1*HEXDIG "." 1*( unreserved / sub-delims / ":" )
'''
IPVFUTURE_RE = r'(v{hexdig}{{1}}\.({unreserved}|{sub_delims}|\:){{1}})'.format(
    hexdig=HEXDIG_RE,
    unreserved=UNRESERVED_RE,
    sub_delims=SUB_DELIMS_RE
    )

'''
   IP-literal    = "[" ( IPv6address / IPvFuture  ) "]"
'''
IP_LITERAL_RE = r'(({ipv6address}|{ipvfuture}))?'.format(
    ipv6address=IPV6ADDRESS_RE,
    ipvfuture=IPVFUTURE_RE
    )

'''
   scheme        = ALPHA *( ALPHA / DIGIT / "+" / "-" / "." )
'''
SCHEME_RE = r'({alpha}({alpha}|{digit}|\+|\-|\.)*)'.format(
    alpha=ALPHA_RE,
    digit=DIGIT_RE
    )

'''
   userinfo      = *( unreserved / pct-encoded / sub-delims / ":" )
'''
USERINFO_RE = r'(({unreserved}|{pct_encoded}|{sub_delims}|\:)*)'.format(
    unreserved=UNRESERVED_RE,
    pct_encoded=PCT_ENCODED_RE,
    sub_delims=SUB_DELIMS_RE
    )

'''
   host          = IP-literal / IPv4address / reg-name
'''
HOST_RE = r'({ip_literal}|{ipv4address}|{reg_name})'.format(
    ip_literal=IP_LITERAL_RE,
    ipv4address=IPV4ADDRESS_RE,
    reg_name=REG_NAME_RE
)

'''
   port          = *DIGIT
'''
PORT_RE = r'{digit}*'.format(digit=DIGIT_RE)

'''
   authority     = [ userinfo "@" ] host [ ":" port ]
'''
AUTHORITY_RE = r'(({userinfo}\@)?{host}(\:{port})?)'.format(
    userinfo=USERINFO_RE,
    host=HOST_RE,
    port=PORT_RE
    )

'''
   relative-part = "//" authority path-abempty
                 / path-absolute
                 / path-noscheme
                 / path-empty
'''
RELATIVE_PART_RE = (
    r'((\/\/{authority}{path_abempty})|'
    r'{path_absolute}|{path_noscheme}|{path_empty})'
    ).format(
        authority=AUTHORITY_RE,
        path_abempty=PATH_ABEMPTY_RE,
        path_absolute=PATH_ABSOLUTE_RE,
        path_noscheme=PATH_NOSCHEME_RE,
        path_empty=PATH_EMPTY_RE
        )

'''
   relative-ref  = relative-part [ "?" query ] [ "#" fragment ]
'''
RELATIVE_REF_RE = r'({relative_part}(\?{query})?(\#{fragment})?)'.format(
    relative_part=RELATIVE_PART_RE,
    query=QUERY_RE,
    fragment=FRAGMENT_RE
    )

'''
   hier-part     = "//" authority path-abempty
                 / path-absolute
                 / path-rootless
                 / path-empty
'''
HIER_PART_RE = (
    r'((\/\/{authority}{path_abempty})|'
    r'{path_absolute}|{path_rootless}|{path_empty})'
    ).format(
        authority=AUTHORITY_RE,
        path_abempty=PATH_ABEMPTY_RE,
        path_absolute=PATH_ABSOLUTE_RE,
        path_rootless=PATH_ROOTLESS_RE,
        path_empty=PATH_EMPTY_RE
        )

'''
   URI           = scheme ":" hier-part [ "?" query ] [ "#" fragment ]
'''
URI_RE = r'({scheme}\:{hier_part}(\?{query})?(\#{fragment})?)'.format(
    scheme=SCHEME_RE,
    hier_part=HIER_PART_RE,
    query=QUERY_RE,
    fragment=FRAGMENT_RE
    )

'''
   URI-reference = URI / relative-ref
'''
URI_REFERENCE_RE = r'({uri}|{relative_ref})'.format(
    uri=URI_RE,
    relative_ref=RELATIVE_REF_RE
    )

'''
   absolute-URI  = scheme ":" hier-part [ "?" query ]
'''
ABSOLUTE_URI_RE = r'({scheme}\:{hier_part}(\?{query})?)'.format(
    scheme=SCHEME_RE,
    hier_part=HIER_PART_RE,
    query=QUERY_RE
    )




import copy
import mmap
import os.path
import logging

from wayround_org.utils.format.elf_enum import *

from wayround_org.utils.format.elf_bin import (
    read_e_ident,
    is_elf,
    e_ident_bitness,
    e_ident_endianness,
    e_ident_to_dict,

    read_elf_ehdr_x,
    read_elf_ehdr,

    elf32_ehdr_to_dict,
    elf64_ehdr_to_dict,
    elf_ehdr_to_dict,

    read_elf_shdr_x,
    read_elf_shdr,

    read_elf_phdr_x,
    read_elf_phdr,


    read_elf_section_header_table,
    read_elf_section_header_table_names,

    read_elf_program_header_table,

    read_dynamic_section,

    get_dynamic_libs_names,
    get_dynamic_runpath_values,
    convert_virtual_to_file
    )


def dict_byte_to_ints(elf_ehdr_dict, only_keys=None, endianness='little'):

    ret = copy.copy(elf_ehdr_dict)

    for i in ret.keys():
        work_it = False

        if isinstance(only_keys, (list, set)):
            if i in only_keys:
                work_it = True
        else:
            work_it = True

        if work_it:
            ret[i] = int.from_bytes(ret[i], endianness)

    return ret


def e_ident_text(e_ident_dict):

    ret = """\
Class:       {e_i_s_class}
Data:        {e_i_s_data}
Version:     {e_i_s_version}
OS ABI:      {e_i_s_osabi}
ABI Version: {e_i_s_abiversion}
""".format_map(e_ident_dict)

    return ret


def elf_x_ehdr_text(elf_ehdr_dict, endianness):

    ret = """\
Object file type                     0x{e_type:x}
Architecture                         0x{e_machine:x}
Object file version                  {e_version}
Entry point virtual address          0x{e_entry:08x}
Program header table file offset     0x{e_phoff:08x} ({e_phoff} B)
Section header table file offset     0x{e_shoff:08x} ({e_shoff} B)
Processor-specific flags             b{e_flags:b}
ELF header size in bytes             {e_ehsize} B
Program header table entry size      {e_phentsize} B
Program header table entry count     {e_phnum}
Section header table entry size      {e_shentsize} B
Section header table entry count     {e_shnum}
Section header string table index    {e_shstrndx}
""".format_map(dict_byte_to_ints(elf_ehdr_dict, endianness=endianness))

    return ret


def get_section_header_type_name(value):

    ret = None

    names = {
        SHT_NULL: 'NULL',
        SHT_PROGBITS: 'PROGBITS',
        SHT_SYMTAB: 'SYMTAB',
        SHT_STRTAB: 'STRTAB',
        SHT_RELA: 'RELA',
        SHT_HASH: 'HASH',
        SHT_DYNAMIC: 'DYNAMIC',
        SHT_NOTE: 'NOTE',
        SHT_NOBITS: 'NOBITS',
        SHT_REL: 'REL',
        SHT_SHLIB: 'SHLIB',
        SHT_DYNSYM: 'DYNSYM',
        SHT_INIT_ARRAY: 'INIT_ARRAY',
        SHT_FINI_ARRAY: 'FINI_ARRAY',
        SHT_PREINIT_ARRAY: 'PREINIT_ARRAY',
        SHT_GROUP: 'GROUP',
        SHT_SYMTAB_SHNDX: 'SYMTAB_SHNDX',
        SHT_NUM: 'NUM',
        SHT_LOOS: 'LOOS',
        SHT_GNU_ATTRIBUTES: 'GNU_ATTRIBUTES',
        SHT_GNU_HASH: 'GNU_HASH',
        SHT_GNU_LIBLIST: 'GNU_LIBLIST',
        SHT_CHECKSUM: 'CHECKSUM',
        SHT_LOSUNW: 'LOSUNW',
        SHT_SUNW_move: 'SUNW_move',
        SHT_SUNW_COMDAT: 'SUNW_COMDAT',
        SHT_SUNW_syminfo: 'SUNW_syminfo',
        SHT_GNU_verdef: 'GNU_verdef',
        SHT_GNU_verneed: 'GNU_verneed',
        SHT_GNU_versym: 'GNU_versym',
        SHT_HISUNW: 'HISUNW',
        SHT_HIOS: 'HIOS',
        SHT_LOPROC: 'LOPROC',
        SHT_HIPROC: 'HIPROC',
        SHT_LOUSER: 'LOUSER',
        SHT_HIUSER: 'HIUSER'
        }

    if value in names:
        ret = names[value]
    else:
        ret = '0x{:x}'.format(value)

    return ret


def get_program_header_type_name(value):

    ret = None

    names = {
        PT_NULL: 'NULL',
        PT_LOAD: 'LOAD',
        PT_DYNAMIC: 'DYNAMIC',
        PT_INTERP: 'INTERP',
        PT_NOTE: 'NOTE',
        PT_SHLIB: 'SHLIB',
        PT_PHDR: 'PHDR',
        PT_TLS: 'TLS',
        PT_NUM: 'NUM',
        PT_LOOS: 'LOOS',
        PT_GNU_EH_FRAME: 'GNU_EH_FRAME',
        PT_GNU_STACK: 'GNU_STACK',
        PT_GNU_RELRO: 'GNU_RELRO',
        PT_LOSUNW: 'LOSUNW',
        PT_SUNWBSS: 'SUNWBSS',
        PT_SUNWSTACK: 'SUNWSTACK',
        PT_HISUNW: 'HISUNW',
        PT_HIOS: 'HIOS',
        PT_LOPROC: 'LOPROC',
        PT_HIPROC: 'HIPROC'
        }

    if value in names:
        ret = names[value]
    else:
        ret = '0x{:x}'.format(value)

    return ret


def get_dynamic_type_name(value):
    ret = None

    names = {
        DT_NULL: 'DT_NULL',
        DT_NEEDED: 'DT_NEEDED',
        DT_PLTRELSZ: 'DT_PLTRELSZ',
        DT_PLTGOT: 'DT_PLTGOT',
        DT_HASH: 'DT_HASH',
        DT_STRTAB: 'DT_STRTAB',
        DT_SYMTAB: 'DT_SYMTAB',
        DT_RELA: 'DT_RELA',
        DT_RELASZ: 'DT_RELASZ',
        DT_RELAENT: 'DT_RELAENT',
        DT_STRSZ: 'DT_STRSZ',
        DT_SYMENT: 'DT_SYMENT',
        DT_INIT: 'DT_INIT',
        DT_FINI: 'DT_FINI',
        DT_SONAME: 'DT_SONAME',
        DT_RPATH: 'DT_RPATH',
        DT_SYMBOLIC: 'DT_SYMBOLIC',
        DT_REL: 'DT_REL',
        DT_RELSZ: 'DT_RELSZ',
        DT_RELENT: 'DT_RELENT',
        DT_PLTREL: 'DT_PLTREL',
        DT_DEBUG: 'DT_DEBUG',
        DT_TEXTREL: 'DT_TEXTREL',
        DT_JMPREL: 'DT_JMPREL',
        DT_BIND_NOW: 'DT_BIND_NOW',
        DT_INIT_ARRAY: 'DT_INIT_ARRAY',
        DT_FINI_ARRAY: 'DT_FINI_ARRAY',
        DT_INIT_ARRAYSZ: 'DT_INIT_ARRAYSZ',
        DT_FINI_ARRAYSZ: 'DT_FINI_ARRAYSZ',
        DT_RUNPATH: 'DT_RUNPATH',
        DT_FLAGS: 'DT_FLAGS',
        DT_ENCODING: 'DT_ENCODING',
        DT_PREINIT_ARRAY: 'DT_PREINIT_ARRAY',
        DT_PREINIT_ARRAYSZ: 'DT_PREINIT_ARRAYSZ',
        DT_NUM: 'DT_NUM',
        DT_LOOS: 'DT_LOOS',
        DT_HIOS: 'DT_HIOS',
        DT_LOPROC: 'DT_LOPROC',
        DT_HIPROC: 'DT_HIPROC',
        DT_PROCNUM: 'DT_PROCNUM',
        DT_VALRNGLO: 'DT_VALRNGLO',
        DT_GNU_PRELINKED: 'DT_GNU_PRELINKED',
        DT_GNU_CONFLICTSZ: 'DT_GNU_CONFLICTSZ',
        DT_GNU_LIBLISTSZ: 'DT_GNU_LIBLISTSZ',
        DT_CHECKSUM: 'DT_CHECKSUM',
        DT_PLTPADSZ: 'DT_PLTPADSZ',
        DT_MOVEENT: 'DT_MOVEENT',
        DT_MOVESZ: 'DT_MOVESZ',
        DT_FEATURE_1: 'DT_FEATURE_1',
        DT_POSFLAG_1: 'DT_POSFLAG_1',
        DT_SYMINSZ: 'DT_SYMINSZ',
        DT_SYMINENT: 'DT_SYMINENT',
        DT_VALRNGHI: 'DT_VALRNGHI',
        DT_VALNUM: 'DT_VALNUM',
        DT_ADDRRNGLO: 'DT_ADDRRNGLO',
        DT_GNU_HASH: 'DT_GNU_HASH',
        DT_TLSDESC_PLT: 'DT_TLSDESC_PLT',
        DT_TLSDESC_GOT: 'DT_TLSDESC_GOT',
        DT_GNU_CONFLICT: 'DT_GNU_CONFLICT',
        DT_GNU_LIBLIST: 'DT_GNU_LIBLIST',
        DT_CONFIG: 'DT_CONFIG',
        DT_DEPAUDIT: 'DT_DEPAUDIT',
        DT_AUDIT: 'DT_AUDIT',
        DT_PLTPAD: 'DT_PLTPAD',
        DT_MOVETAB: 'DT_MOVETAB',
        DT_SYMINFO: 'DT_SYMINFO',
        DT_ADDRRNGHI: 'DT_ADDRRNGHI',
        DT_ADDRNUM: 'DT_ADDRNUM',
        DT_VERSYM: 'DT_VERSYM',
        DT_RELACOUNT: 'DT_RELACOUNT',
        DT_RELCOUNT: 'DT_RELCOUNT',
        DT_FLAGS_1: 'DT_FLAGS_1',
        DT_VERDEF: 'DT_VERDEF',
        DT_VERDEFNUM: 'DT_VERDEFNUM',
        DT_VERNEED: 'DT_VERNEED',
        DT_VERNEEDNUM: 'DT_VERNEEDNUM',
        DT_VERSIONTAGNUM: 'DT_VERSIONTAGNUM',
        DT_AUXILIARY: 'DT_AUXILIARY',
        DT_FILTER: 'DT_FILTER',
        DT_EXTRANUM: 'DT_EXTRANUM',
        }

    if value in names:
        ret = names[value]
    else:
        ret = '0x{:x}'.format(value)

    return ret

MACHINE_DICT = {
    EM_NONE: {
        'name': 'EM_NONE',
        'descr': 'No machine'
    },
    EM_M32: {
        'name': 'EM_M32',
        'descr': 'AT&T WE 32100'
    },
    EM_SPARC: {
        'name': 'EM_SPARC',
        'descr': 'SUN SPARC'
    },
    EM_386: {
        'name': 'EM_386',
        'descr': 'Intel 80386'
    },
    EM_68K: {
        'name': 'EM_68K',
        'descr': 'Motorola m68k family'
    },
    EM_88K: {
        'name': 'EM_88K',
        'descr': 'Motorola m88k family'
    },
    EM_860: {
        'name': 'EM_860',
        'descr': 'Intel 80860'
    },
    EM_MIPS: {
        'name': 'EM_MIPS',
        'descr': 'MIPS R3000 big-endian'
    },
    EM_S370: {
        'name': 'EM_S370',
        'descr': 'IBM System/370'
    },
    EM_MIPS_RS3_LE: {
        'name': 'EM_MIPS_RS3_LE',
        'descr': 'MIPS R3000 little-endian'
    },
    EM_PARISC: {
        'name': 'EM_PARISC',
        'descr': 'HPPA'
    },
    EM_VPP500: {
        'name': 'EM_VPP500',
        'descr': 'Fujitsu VPP500'
    },
    EM_SPARC32PLUS: {
        'name': 'EM_SPARC32PLUS',
        'descr': 'Sun\'s "v8plus"'
    },
    EM_960: {
        'name': 'EM_960',
        'descr': 'Intel 80960'
    },
    EM_PPC: {
        'name': 'EM_PPC',
        'descr': 'PowerPC'
    },
    EM_PPC64: {
        'name': 'EM_PPC64',
        'descr': 'PowerPC 64-bit'
    },
    EM_S390: {
        'name': 'EM_S390',
        'descr': 'IBM S390'
    },
    EM_V800: {
        'name': 'EM_V800',
        'descr': 'NEC V800 series'
    },
    EM_FR20: {
        'name': 'EM_FR20',
        'descr': 'Fujitsu FR20'
    },
    EM_RH32: {
        'name': 'EM_RH32',
        'descr': 'TRW RH-32'
    },
    EM_RCE: {
        'name': 'EM_RCE',
        'descr': 'Motorola RCE'
    },
    EM_ARM: {
        'name': 'EM_ARM',
        'descr': 'ARM'
    },
    EM_FAKE_ALPHA: {
        'name': 'EM_FAKE_ALPHA',
        'descr': 'Digital Alpha'
    },
    EM_SH: {
        'name': 'EM_SH',
        'descr': 'Hitachi SH'
    },
    EM_SPARCV9: {
        'name': 'EM_SPARCV9',
        'descr': 'SPARC v9 64-bit'
    },
    EM_TRICORE: {
        'name': 'EM_TRICORE',
        'descr': 'Siemens Tricore'
    },
    EM_ARC: {
        'name': 'EM_ARC',
        'descr': 'Argonaut RISC Core'
    },
    EM_H8_300: {
        'name': 'EM_H8_300',
        'descr': 'Hitachi H8/300'
    },
    EM_H8_300H: {
        'name': 'EM_H8_300H',
        'descr': 'Hitachi H8/300H'
    },
    EM_H8S: {
        'name': 'EM_H8S',
        'descr': 'Hitachi H8S'
    },
    EM_H8_500: {
        'name': 'EM_H8_500',
        'descr': 'Hitachi H8/500'
    },
    EM_IA_64: {
        'name': 'EM_IA_64',
        'descr': 'Intel Merced'
    },
    EM_MIPS_X: {
        'name': 'EM_MIPS_X',
        'descr': 'Stanford MIPS-X'
    },
    EM_COLDFIRE: {
        'name': 'EM_COLDFIRE',
        'descr': 'Motorola Coldfire'
    },
    EM_68HC12: {
        'name': 'EM_68HC12',
        'descr': 'Motorola M68HC12'
    },
    EM_MMA: {
        'name': 'EM_MMA',
        'descr': 'Fujitsu MMA Multimedia Accelerator'
    },
    EM_PCP: {
        'name': 'EM_PCP',
        'descr': 'Siemens PCP'
    },
    EM_NCPU: {
        'name': 'EM_NCPU',
        'descr': 'Sony nCPU embeeded RISC'
    },
    EM_NDR1: {
        'name': 'EM_NDR1',
        'descr': 'Denso NDR1 microprocessor'
    },
    EM_STARCORE: {
        'name': 'EM_STARCORE',
        'descr': 'Motorola Start*Core processor'
    },
    EM_ME16: {
        'name': 'EM_ME16',
        'descr': 'Toyota ME16 processor'
    },
    EM_ST100: {
        'name': 'EM_ST100',
        'descr': 'STMicroelectronic ST100 processor'
    },
    EM_TINYJ: {
        'name': 'EM_TINYJ',
        'descr': 'Advanced Logic Corp. Tinyj emb.fam'
    },
    EM_X86_64: {
        'name': 'EM_X86_64',
        'descr': 'AMD x86-64 architecture'
    },
    EM_PDSP: {
        'name': 'EM_PDSP',
        'descr': 'Sony DSP Processor'
    },
    EM_FX66: {
        'name': 'EM_FX66',
        'descr': 'Siemens FX66 microcontroller'
    },
    EM_ST9PLUS: {
        'name': 'EM_ST9PLUS',
        'descr': 'STMicroelectronics ST9+ 8/16 mc'
    },
    EM_ST7: {
        'name': 'EM_ST7',
        'descr': 'STmicroelectronics ST7 8 bit mc'
    },
    EM_68HC16: {
        'name': 'EM_68HC16',
        'descr': 'Motorola MC68HC16 microcontroller'
    },
    EM_68HC11: {
        'name': 'EM_68HC11',
        'descr': 'Motorola MC68HC11 microcontroller'
    },
    EM_68HC08: {
        'name': 'EM_68HC08',
        'descr': 'Motorola MC68HC08 microcontroller'
    },
    EM_68HC05: {
        'name': 'EM_68HC05',
        'descr': 'Motorola MC68HC05 microcontroller'
    },
    EM_SVX: {
        'name': 'EM_SVX',
        'descr': 'Silicon Graphics SVx'
    },
    EM_ST19: {
        'name': 'EM_ST19',
        'descr': 'STMicroelectronics ST19 8 bit mc'
    },
    EM_VAX: {
        'name': 'EM_VAX',
        'descr': 'Digital VAX'
    },
    EM_CRIS: {
        'name': 'EM_CRIS',
        'descr': 'Axis Communications 32-bit embedded processor'
    },
    EM_JAVELIN: {
        'name': 'EM_JAVELIN',
        'descr': 'Infineon Technologies 32-bit embedded processor'
    },
    EM_FIREPATH: {
        'name': 'EM_FIREPATH',
        'descr': 'Element 14 64-bit DSP Processor'
    },
    EM_ZSP: {
        'name': 'EM_ZSP',
        'descr': 'LSI Logic 16-bit DSP Processor'
    },
    EM_MMIX: {
        'name': 'EM_MMIX',
        'descr': 'Donald Knuth\'s educational 64 - bit processor'
    },
    EM_HUANY: {
        'name': 'EM_HUANY',
        'descr': 'Harvard University machine-independent object files'
    },
    EM_PRISM: {
        'name': 'EM_PRISM',
        'descr': 'SiTera Prism'
    },
    EM_AVR: {
        'name': 'EM_AVR',
        'descr': 'Atmel AVR 8-bit microcontroller'
    },
    EM_FR30: {
        'name': 'EM_FR30',
        'descr': 'Fujitsu FR30'
    },
    EM_D10V: {
        'name': 'EM_D10V',
        'descr': 'Mitsubishi D10V'
    },
    EM_D30V: {
        'name': 'EM_D30V',
        'descr': 'Mitsubishi D30V'
    },
    EM_V850: {
        'name': 'EM_V850',
        'descr': 'NEC v850'
    },
    EM_M32R: {
        'name': 'EM_M32R',
        'descr': 'Mitsubishi M32R'
    },
    EM_MN10300: {
        'name': 'EM_MN10300',
        'descr': 'Matsushita MN10300'
    },
    EM_MN10200: {
        'name': 'EM_MN10200',
        'descr': 'Matsushita MN10200'
    },
    EM_PJ: {
        'name': 'EM_PJ',
        'descr': 'picoJava'
    },
    EM_OPENRISC: {
        'name': 'EM_OPENRISC',
        'descr': 'OpenRISC 32-bit embedded processor'
    },
    EM_ARC_A5: {
        'name': 'EM_ARC_A5',
        'descr': 'ARC Cores Tangent-A5'
    },
    EM_XTENSA: {
        'name': 'EM_XTENSA',
        'descr': 'Tensilica Xtensa Architecture'
    },
    EM_TILEPRO: {
        'name': 'EM_TILEPRO',
        'descr': 'Tilera TILEPro'
    },
    EM_TILEGX: {
        'name': 'EM_TILEGX',
        'descr': 'Tilera TILE-Gx'
    },
    EM_NUM: {
        'name': 'EM_NUM',
        'descr': ''
    },
    EM_ALPHA: {
        'name': 'EM_ALPHA',
        'descr': ''
    }
}


def machine_name_by_int(value):
    return MACHINE_DICT[value]['name']


def machine_descr_by_int(value):
    return MACHINE_DICT[value]['descr']


def machine_int_by_name(name):
    ret = None
    for i, j in MACHINE_DICT:
        if j['name'] == name:
            ret = i
            break
    return ret


def section_header_table_text(
        data,
        elf_section_header_table,
        ehdr_dict,
        endianness,
        bitness
        ):

    ret = ''

    names = read_elf_section_header_table_names(
        data, elf_section_header_table, ehdr_dict, endianness
        )

    section_header_table2 = []
    for i in elf_section_header_table:
        section_header_table2.append(
            dict_byte_to_ints(
                i,
                endianness=endianness
                )
            )

    elf_section_header_table = section_header_table2

    longest = 0
    for i in names:
        if len(i) > longest:
            longest = len(i)

    types = []
    for i in range(len(elf_section_header_table)):
        types.append(
            get_section_header_type_name(
                elf_section_header_table[i]['sh_type'])
            )

    longest_t = 0
    for i in types:
        if len(i) > longest_t:
            longest_t = len(i)

    if bitness == 32:

        ret += "  [{index:2}] "\
            "{name}(sto:{name_addr:5}) "\
            "{typ} "\
            "{addr:8} "\
            "{off:8} "\
            "{size:8} "\
            "{es:2} "\
            "{flg:10} "\
            "{lk:3} "\
            "{inf:3} "\
            "{al:3}\n".format(
                index='No',
                name='Name'.ljust(longest),
                name_addr='',
                typ='Type'.ljust(longest_t),
                flg='Flags',
                addr='Address',
                off='Offset',
                size='Size',
                lk='Lnk',
                inf='Inf',
                al='Al',
                es='SZ'
            )

        for i in range(len(elf_section_header_table)):
            ret += "  [{index:2}] "\
                "{name}(sto:{name_addr:5x}) "\
                "{typ} "\
                "{addr:08x} "\
                "{off:08x} "\
                "{size:08x} "\
                "{es:02x} "\
                "{flg:010b} "\
                "{lk:03x} "\
                "{inf:03x} "\
                "{al:03x}\n".format(
                    index=i,
                    name=names[i].ljust(longest),
                    name_addr=elf_section_header_table[i]['sh_name'],
                    typ=types[i].ljust(longest_t),
                    flg=elf_section_header_table[i]['sh_flags'],
                    addr=elf_section_header_table[i]['sh_addr'],
                    off=elf_section_header_table[i]['sh_offset'],
                    size=elf_section_header_table[i]['sh_size'],
                    lk=elf_section_header_table[i]['sh_link'],
                    inf=elf_section_header_table[i]['sh_info'],
                    al=elf_section_header_table[i]['sh_addralign'],
                    es=elf_section_header_table[i]['sh_entsize']
                )

    elif bitness == 64:

        ret += "  [{index:2}] "\
            "{name}(sto:{name_addr:5}) "\
            "{typ} "\
            "{addr:16} "\
            "{off:16} "\
            "{size:16} "\
            "{es:2} "\
            "{flg:10} "\
            "{lk:3} "\
            "{inf:3} "\
            "{al:3}\n".format(
                index='No',
                name='Name'.ljust(longest),
                name_addr='',
                typ='Type'.ljust(longest_t),
                flg='Flags',
                addr='Address',
                off='Offset',
                size='Size',
                lk='Lnk',
                inf='Inf',
                al='Al',
                es='SZ'
            )

        for i in range(len(elf_section_header_table)):
            ret += "  [{index:2}] "\
                "{name}(sto:{name_addr:5x}) "\
                "{typ} "\
                "{addr:016x} "\
                "{off:016x} "\
                "{size:016x} "\
                "{es:02x} "\
                "{flg:010b} "\
                "{lk:03x} "\
                "{inf:03x} "\
                "{al:03x}\n".format(
                    index=i,
                    name=names[i].ljust(longest),
                    name_addr=elf_section_header_table[i]['sh_name'],
                    typ=types[i].ljust(longest_t),
                    flg=elf_section_header_table[i]['sh_flags'],
                    addr=elf_section_header_table[i]['sh_addr'],
                    off=elf_section_header_table[i]['sh_offset'],
                    size=elf_section_header_table[i]['sh_size'],
                    lk=elf_section_header_table[i]['sh_link'],
                    inf=elf_section_header_table[i]['sh_info'],
                    al=elf_section_header_table[i]['sh_addralign'],
                    es=elf_section_header_table[i]['sh_entsize']
                )

    else:
        raise Exception("unknown bitness: {}".format(bitness))

    ret += '\n'
    ret += 'sto - string table offset\n'

    return ret


def program_header_table_text(program_header_table, endianness, bitness=32):

    ret = ''

    program_header_table2 = []
    for i in program_header_table:
        program_header_table2.append(
            dict_byte_to_ints(
                i,
                endianness=endianness
                )
            )

    program_header_table = program_header_table2

    types = []
    for i in range(len(program_header_table)):
        types.append(
            get_program_header_type_name(
                program_header_table[i]['p_type']))

    longest_t = 0
    for i in types:
        if len(i) > longest_t:
            longest_t = len(i)

    if bitness == 32:

        ret += "  {typ} "\
            "{offset:8} "\
            "{virtaddr:8} "\
            "{physaddr:8} "\
            "{filesize:8} "\
            "{memsize:8} "\
            "{flag:10} "\
            "{align:8}\n".format(
                typ='Type'.ljust(longest_t),
                flag='Flags',
                offset='F Offset',
                virtaddr='VirtAddr',
                physaddr='PhisAddr',
                filesize='FileSize',
                memsize='MemSize',
                align='Align'
            )

        for i in range(len(program_header_table)):
            ret += "  {typ} "\
                "{offset:08x} "\
                "{virtaddr:08x} "\
                "{physaddr:08x} "\
                "{filesize:08x} "\
                "{memsize:08x} "\
                "{flag:010b} "\
                "{align:08x}\n".format(
                    typ=types[i].ljust(longest_t),
                    flag=program_header_table[i]['p_flags'],
                    offset=program_header_table[i]['p_offset'],
                    virtaddr=program_header_table[i]['p_vaddr'],
                    physaddr=program_header_table[i]['p_paddr'],
                    filesize=program_header_table[i]['p_filesz'],
                    memsize=program_header_table[i]['p_memsz'],
                    align=program_header_table[i]['p_align']
                )

    elif bitness == 64:

        ret += "  {typ} "\
            "{offset:16} "\
            "{virtaddr:16} "\
            "{physaddr:16} "\
            "{filesize:16} "\
            "{memsize:16} "\
            "{flag:10} "\
            "{align:16}\n".format(
                typ='Type'.ljust(longest_t),
                flag='Flags',
                offset='F Offset',
                virtaddr='VirtAddr',
                physaddr='PhisAddr',
                filesize='FileSize',
                memsize='MemSize',
                align='Align'
            )

        for i in range(len(program_header_table)):
            ret += "  {typ} "\
                "{offset:016x} "\
                "{virtaddr:016x} "\
                "{physaddr:016x} "\
                "{filesize:016x} "\
                "{memsize:016x} "\
                "{flag:010b} "\
                "{align:016x}\n".format(
                    typ=types[i].ljust(longest_t),
                    flag=program_header_table[i]['p_flags'],
                    offset=program_header_table[i]['p_offset'],
                    virtaddr=program_header_table[i]['p_vaddr'],
                    physaddr=program_header_table[i]['p_paddr'],
                    filesize=program_header_table[i]['p_filesz'],
                    memsize=program_header_table[i]['p_memsz'],
                    align=program_header_table[i]['p_align']
                )

    else:
        raise Exception("unknown bitness: {}".format(bitness))

    return ret


def dynamic_section_text(dinamics_table, endianness):

    ret = ''

    names = []

    for i in range(len(dinamics_table)):
        names.append(get_dynamic_type_name(dinamics_table[i]['d_tag']))

    longest_t = 0
    for i in names:
        if len(i) > longest_t:
            longest_t = len(i)

    for i in dinamics_table:
        number = 0
        if get_dynamic_type_name(i['d_tag']) == 'DT_NEEDED':
            number = i['d_ptr']
        else:
            number = i['d_val']
        ret += "{}  {:08x}\n".format(
            get_dynamic_type_name(i['d_tag']).ljust(longest_t),
            number
            )

    return ret


class ELF:

    def __init__(self, filename, verbose=False, debug=False, mute=False):

        error = False

        self.debug = debug

        self.verbose = verbose

        self.mute = mute

        _FILLABLE_ATTRIBUTE_NAMES = [
            'e_ident',
            'bitness',
            'endianness',
            'e_ident_dict',
            'e_ident_text',
            'elf_ehdr',
            'elf_ehdr_dict',
            'elf_x_ehdr_text',

            'elf_type',
            'elf_type_name',

            'elf_machine',
            'elf_machine_name',
            'elf_machine_descr',

            'section_table',
            'section_names',
            'program_table',
            'section_header_table_text',
            'program_header_table_text',
            'dynamic_section_index',
            'dynamic_section_offset',
            'dynamic_section',
            'dynamic_section_text',
            'needed_libs_list',
            'runpath_values',
            'libs_list_text'
            ]

        if debug or verbose:
            print("File `{}'".format(filename))

        self.opened = False

        self.filename = filename

        if not os.path.isfile(filename):
            if not mute:
                logging.error("Not a file: `{}'".format(filename))
                error = True

        if not error:

            if os.stat(filename).st_size == 0:
                if not mute and verbose:
                    logging.info("File size is 0: {}".format(filename))
                error = True

        f = None
        m = None

        if not error:

            try:
                f = open(filename, 'rb')
            except KeyboardInterrupt:
                error = True
                raise
            except:
                if not mute:
                    logging.exception(
                        "Couldn't open file for read: `{}'".format(
                            filename)
                        )
                error = True

        if not error:

            try:
                m = mmap.mmap(
                    f.fileno(),
                    0,
                    flags=mmap.MAP_PRIVATE,
                    prot=mmap.PROT_READ
                    )
            except KeyboardInterrupt:
                raise
                error = True
            except:
                if not mute:
                    logging.exception(
                        "Couldn't map file: `{}'".format(filename)
                        )
                error = True

        if not error:

            self.opened = True

            self.is_elf = self._fill_is_elf(m, debug)

            errors = 0

            for i in _FILLABLE_ATTRIBUTE_NAMES:

                setattr(self, i, None)

                if self.is_elf:

                    filler = getattr(
                        self,
                        '_fill_{}'.format(i)
                        )

                    filler_res = filler(m, debug)

                    setattr(self, i, filler_res)

                else:
                    # NOTE: not an error. but attributes must
                    #       be set atleast to None, so not a break
                    pass

            if errors != 0:
                error = True

        if m:
            m.close()

        if f:
            f.close()

        self.error = error

        if error:
            self.is_elf = False
            for i in _FILLABLE_ATTRIBUTE_NAMES:
                setattr(self, i, None)

        return

    def return_text(self):
        ret = """\
file: {filename}

is elf?: {is_elf}

({bitness}-bit)
({endianness}-byteorder)

{e_ident_text}

{elf_x_ehdr_text}

Section table:
{section_header_table_text}

Program table:
{program_header_table_text}

Dynamic section:
{dynamic_section_text}

Needed libs list:
{libs_list_text}
""".format(
            filename=self.filename,
            is_elf=self.is_elf,
            bitness=self.bitness,
            endianness=self.endianness,
            e_ident_text=self.e_ident_text,
            elf_x_ehdr_text=self.elf_x_ehdr_text,
            section_header_table_text=self.section_header_table_text,
            program_header_table_text=self.program_header_table_text,
            dynamic_section_text=self.dynamic_section_text,
            libs_list_text=self.libs_list_text
            )

        return ret

    def _fill_is_elf(self, m, debug):
        return is_elf(m)

    def _fill_e_ident(self, m, debug):

        ret = read_e_ident(m)
        if debug:
            print("e_ident == {}".format(ret))

        return ret

    def _fill_bitness(self, m, debug):
        ret = None

        if self.e_ident:
            ret = e_ident_bitness(self.e_ident)

        if debug:
            print("bitness == {}".format(ret))

        return ret

    def _fill_endianness(self, m, debug):

        ret = None
        if self.e_ident:
            ret = e_ident_endianness(self.e_ident)

        if debug:
            print("endianness == {}".format(ret))

        return ret

    def _fill_elf_machine(self, m, debug):

        ret = None
        if self.elf_ehdr_dict and self.endianness:

            ret = int.from_bytes(
                self.elf_ehdr_dict['e_machine'],
                self.endianness
                )

        if debug:
            print("elf_machine == {}".format(ret))

        return ret

    def _fill_elf_machine_name(self, m, debug):
        ret = None

        if self.elf_machine:
            ret = machine_name_by_int(self.elf_machine)

        if debug:
            print("elf_machine_name == {}".format(ret))

        return ret

    def _fill_elf_machine_descr(self, m, debug):

        ret = None

        if self.elf_machine:

            ret = machine_descr_by_int(self.elf_machine)

        if debug:
            print("elf_machine_descr == {}".format(ret))

        return ret

    def _fill_elf_type(self, m, debug):
        ret = None
        if self.elf_ehdr_dict and self.endianness:

            ret = int.from_bytes(
                self.elf_ehdr_dict['e_type'],
                self.endianness
                )

        if debug:
            print("elf_type == {}".format(ret))

        return ret

    def _fill_elf_type_name(self, m, debug):

        ret = 0

        if self.elf_type:

            for i in [
                    'ET_NONE',
                    'ET_REL',
                    'ET_EXEC',
                    'ET_DYN',
                    'ET_CORE',
                    'ET_NUM',
                    'ET_LOOS',
                    'ET_HIOS',
                    'ET_LOPROC',
                    'ET_HIPROC'
                    ]:

                if self.elf_type == eval(i):
                    ret = i
                    break

        if debug:
            print("elf_type_name == {}".format(ret))
        return ret

    def _fill_e_ident_dict(self, m, debug):

        ret = None

        if self.e_ident:
            ret = e_ident_to_dict(self.e_ident)

        if debug:
            print("e_ident_dict == {}".format(ret))

        return ret

    def _fill_e_ident_text(self, m, debug):

        ret = 0

        if self.e_ident_dict:

            ret = e_ident_text(self.e_ident_dict)

        if debug:
            print("e_ident_text == {}".format(ret))

        return ret

    def _fill_elf_ehdr(self, m, debug):

        ret = None

        if self.e_ident_dict:
            ret = read_elf_ehdr(
                m,
                0,
                self.e_ident_dict
                )

        if debug:
            print("elf_ehdr == {}".format(ret))

        return ret

    def _fill_elf_ehdr_dict(self, m, debug):

        ret = None

        if self.e_ident_dict:
            ret = elf_ehdr_to_dict(
                m,
                0,
                self.e_ident_dict
                )

        if debug:
            print("elf_ehdr_dict == {}".format(ret))

        return ret

    def _fill_elf_x_ehdr_text(self, m, debug):

        ret = None

        if self.elf_ehdr_dict and self.endianness:

            ret = elf_x_ehdr_text(
                self.elf_ehdr_dict,
                self.endianness
                )

        if debug:
            print("elf_x_ehdr_text == {}".format(ret))

        return ret

    def _fill_section_table(self, m, debug):

        ret = None

        if self.e_ident_dict and self.elf_ehdr_dict:

            ret = read_elf_section_header_table(
                m,
                self.e_ident_dict,
                self.elf_ehdr_dict
                )

        if debug:
            print("section_table == {}".format(ret))

        return ret

    def _fill_section_names(self, m, debug):

        ret = None

        if (self.section_table
                and self.elf_ehdr_dict
                and self.endianness):

            ret = read_elf_section_header_table_names(
                m,
                self.section_table,
                self.elf_ehdr_dict,
                self.endianness
                )

        if debug:
            print("section_names == {}".format(ret))

        return ret

    def _fill_program_table(self, m, debug):

        ret = None
        if self.e_ident_dict and self.elf_ehdr_dict:
            ret = read_elf_program_header_table(
                m,
                self.e_ident_dict,
                self.elf_ehdr_dict
                )

        if debug:
            print("program_table == {}".format(ret))

        return ret

    def _fill_section_header_table_text(self, m, debug):

        ret = None
        if (self.section_table
                and self.elf_ehdr_dict
                and self.endianness):

            ret = (
                section_header_table_text(
                    m,
                    self.section_table,
                    self.elf_ehdr_dict,
                    self.endianness,
                    self.bitness
                    )
                )

        if debug:
            print("section_header_table_text == {}".format(ret))

        return ret

    def _fill_program_header_table_text(self, m, debug):

        ret = None
        if self.program_table and self.endianness:

            ret = program_header_table_text(
                self.program_table,
                self.endianness,
                self.bitness
                )

        if debug:
            print("program_header_table_text == {}".format(ret))

        return ret

    def _fill_dynamic_section_index(self, m, debug):

        ret = None
        if (
                self.section_names
                and '.dynamic' in self.section_names
                and self.section_table
                and int.from_bytes(
                    self.section_table[
                        self.section_names.index('.dynamic')
                        ]['sh_type'],
                    self.endianness,
                    signed=False
                    ) == SHT_DYNAMIC
                ):

            ret = self.section_names.index('.dynamic')

        if debug:
            print("dynamic_section_index == {}".format(ret))

        return ret

    def _fill_dynamic_section_offset(self, m, debug):

        ret = None
        if (self.section_table
            and self.dynamic_section_index
            and self.endianness
            and 'sh_offset' in self.section_table[
                        self.dynamic_section_index
                        ]
            ):

            ret = int.from_bytes(
                self.section_table[self.dynamic_section_index]['sh_offset'],
                self.endianness
                )

        if debug:
            print("dynamic_section_offset == {}".format(ret))

        return ret

    def _fill_dynamic_section(self, m, debug):

        ret = None
        if self.dynamic_section_offset and self.bitness and self.endianness:
            ret = read_dynamic_section(
                m,
                self.dynamic_section_offset,
                self.bitness,
                self.endianness
                )

        if debug:
            print("dynamic_section == {}".format(ret))

        return ret

    def _fill_dynamic_section_text(self, m, debug):

        ret = None
        if self.dynamic_section and self.endianness:
            ret = dynamic_section_text(
                self.dynamic_section,
                self.endianness
                )

        if debug:
            print("dynamic_section_text == {}".format(ret))
        return ret

    def _fill_needed_libs_list(self, m, debug):

        ret = None
        if (self.program_table
                and self.dynamic_section
                and self.section_table
                and self.endianness):

            ret = get_dynamic_libs_names(
                m,
                self.program_table,
                self.dynamic_section,
                self.section_table,
                self.endianness
                )

        if debug:
            print("needed_libs_list == {}".format(ret))

        return ret

    def _fill_runpath_values(self, m, debug):

        ret = None
        if (self.program_table
                and self.dynamic_section
                and self.section_table
                and self.endianness):

            ret = get_dynamic_runpath_values(
                m,
                self.program_table,
                self.dynamic_section,
                self.section_table,
                self.endianness
                )

        if debug:
            print("runpath_values == {}".format(ret))

        return ret

    def _fill_libs_list_text(self, m, debug):
        ret = None

        if self.needed_libs_list:
            ret = "{}.".format(', '.join(self.needed_libs_list))

        if debug:
            print("libs_list_text == {}".format(ret))

        return ret




import threading
import logging
import os
import io
import sys


import wayround_org.utils.path
import wayround_org.utils.stream
import wayround_org.utils.time
import wayround_org.utils.error
import wayround_org.utils.osutils


class LoggingFileLikeObject:

    def __init__(self, log_instance, typ='info'):

        if not type(log_instance) == Log:
            raise TypeError("only Log instance allowed here")

        if not typ in ['info', 'error']:
            raise ValueError("invalid typ value")

        self._log = log_instance
        self._typ = typ
        self._pipe = os.pipe()
        self._pipe_read_file = os.fdopen(self._pipe[0])
        self._stop_flag = False

        self._close_lock = threading.Lock()

        self._thread = threading.Thread(
            name="Thread of {}".format(self),
            target=self._thread_run
            )

        self._thread.start()

        return

    def fileno(self):
        return self._pipe[1]

    def close(self):
        with self._close_lock:
            ret = os.close(self._pipe[1])
            self._stop_flag = True
        return ret

    def _thread_run(self):
        for line in iter(self._pipe_read_file.readline, ''):

            with self._close_lock:

                if self._stop_flag:
                    break

                line = line.rstrip(' \n\r\0')

                if self._typ == 'info':
                    self._log.info(line)

                if self._typ == 'error':
                    self._log.error(line)

        self._pipe_read_file.close()

        return


def process_output_logger(process, log):

    raise Exception(
        "this is depricated. use .stdout and/or .stderr objects"
        )

    t = wayround_org.utils.stream.lbl_write(
        process.stdout,
        log,
        True
        )
    t.start()

    t2 = wayround_org.utils.stream.lbl_write(
        process.stderr,
        log,
        threaded=True,
        typ='error'
        )
    t2.start()

    t.join()
    t2.join()

    return


class Log:

    def __init__(
            self,
            log_dir,
            logname,
            echo=True,
            timestamp=None,
            longest_logname=None,
            group=None,
            user=None
            ):

        ret = 0
        self.code = 0
        self.fileobj = None
        self.logname = logname
        self.log_filename = None
        self.longest_logname = longest_logname
        self._write_lock = threading.Lock()

        # TODO: group and user parameter's behavior need to be improved
        self._group, self._user = wayround_org.utils.osutils.convert_gid_uid(
            group,
            user
            )

        self.stdout = LoggingFileLikeObject(self, 'info')
        self.stderr = LoggingFileLikeObject(self, 'error')

        os.makedirs(log_dir, exist_ok=True)

        if (not os.path.isdir(log_dir)
                or os.path.islink(log_dir)
            ):
            logging.error(
                "Current file type is not acceptable: {}".format(
                    log_dir
                    )
                )
            ret = 2

        if ret == 0:
            if self._user is not None and self._group is not None:
                os.chown(
                    log_dir,
                    self._user,
                    self._group
                    )

        if ret == 0:

            if timestamp is None:
                timestamp = wayround_org.utils.time.currenttime_stamp_iso8601()

            filename = wayround_org.utils.path.abspath(
                os.path.join(
                    log_dir,
                    "{ts:26} {name}.txt".format(
                        name=logname,
                        ts=timestamp
                        )
                    )
                )

            self.log_filename = filename

            try:
                self.fileobj = open(filename, 'w')
            except:
                logging.exception("Error opening log file")
                ret = 3
            else:
                self.info(
                    "[{}] log started" .format(
                        self.logname
                        ),
                    echo=echo,
                    timestamp=timestamp
                    )

        if ret == 0:
            if self._user is not None and self._group is not None:
                os.chown(
                    self.log_filename,
                    self._user,
                    self._group
                    )

        if ret != 0:
            raise Exception("Exception during Log creation. read above.")

        return

    def __del__(self):
        if self:
            if self.fileobj:
                if not self.fileobj.closed:
                    try:
                        self.stop()
                    except:
                        pass
        return

    def stop(self, echo=True):
        if self.fileobj is None:
            raise Exception("Programming error")

        timestamp = wayround_org.utils.time.currenttime_stamp_iso8601()
        self.info(
            "[{}] log ended" .format(
                self.logname
                ),
            echo=echo,
            timestamp=timestamp
            )

        self.stdout.close()
        self.stderr.close()

        self.fileobj.flush()
        self.fileobj.close()

        return

    close = stop

    def write(self, text, echo=False, typ='info', timestamp=None):

        if not typ in ['info', 'error', 'exception', 'warning']:
            raise ValueError("Wrong `typ' parameter")

        if self.fileobj is None:
            raise Exception("Log output file object is None")

        if timestamp is not None:
            pass
        else:
            timestamp = wayround_org.utils.time.currenttime_stamp_iso8601()

        log_name = self.logname
        if self.longest_logname is not None:
            log_name += ' ' * (self.longest_logname - len(self.logname) - 1)

        icon = 'i'
        if typ == 'info':
            icon = 'i'
        elif typ == 'error':
            icon = 'e'
        elif typ == 'exception':
            icon = 'E'
        elif typ == 'warning':
            icon = 'w'
        else:
            icon = '?'

        msg2 = "[{}] [{:26}] [{}] {}".format(
            icon,
            timestamp,
            log_name,
            text
            )

        msg2_scn = "\033[0;1m[{}] [{:26}] [{}]\033[0m {}".format(
            icon,
            timestamp,
            log_name,
            text
            )

        if icon != 'i':
            msg2_scn = "\033[0;1m[\033[0m\033[0;4m\033[0;5m{}\033[0m\033[0;1m] [{:26}] [{}]\033[0m {}".format(
                icon,
                timestamp,
                log_name,
                text
                )

        with self._write_lock:
            self.fileobj.write(msg2 + '\n')
            self.fileobj.flush()
            if echo:
                sys.stderr.write(msg2_scn + '\n')
                sys.stderr.flush()
        return

    def error(self, text, echo=True, timestamp=None):
        self.write(text, echo=echo, typ='error', timestamp=timestamp)
        return

    def exception(self, text, echo=True, timestamp=None, tb=True):
        # EXCEPTION TEXT:
        ei = wayround_org.utils.error.return_instant_exception_info(
            tb=tb
            )
        ttt = """\
{text}
{ei}
""".format(text=text, ei=ei)

        self.write(ttt, echo=echo, typ='exception', timestamp=timestamp)
        return

    def info(self, text, echo=True, timestamp=None):
        self.write(text, echo=echo, typ='info', timestamp=timestamp)
        return

    def warning(self, text, echo=True, timestamp=None):
        self.write(text, echo=echo, typ='warning', timestamp=timestamp)
        return




"""
Based on RFC3501
"""


import re
import datetime

import wayround_org.utils.datetime_iso8601


DATE_EXPRESSION = (
    r'\s*'
    r'(?P<day>(\s|\d)?\d)'
    r'-'
    r'(?P<month>[a-zA-Z]{3})'
    r'-'
    r'(?P<year>\d+)'
    r'\s*'
    )

DATE_EXPRESSION_C = re.compile(DATE_EXPRESSION)

ZONE_EXPRESSION = (
    r'\s*'
    r'(?P<zone_sign>[+-]?)'
    r'(?P<zone_val>(?P<zone_hours>\d{2})(?P<zone_minutes>\d{2}))'
    r'\s*'
    )

TIME_EXPRESSION = (
    r'\s*'
    r'(?P<hour>\d+)'
    r'\:(?P<minute>\d+)'
    r'\:(?P<second>\d+)'
    r'\s*'
    )

DATETIME_EXPRESSION = (
    r'\s*'
    r'(?P<date>{date})'
    r'\s+'
    r'(?P<time>{time})'
    r'\s+'
    r'(?P<zone>{zone})'
    r'\s*'
    r''
    ).format(
        date=DATE_EXPRESSION,
        time=TIME_EXPRESSION,
        zone=ZONE_EXPRESSION
        )

DATETIME_EXPRESSION_C = re.compile(DATETIME_EXPRESSION)

MONTHES = [
    "Jan", "Feb", "Mar", "Apr",
    "May", "Jun", "Jul", "Aug",
    "Sep", "Oct", "Nov", "Dec"
    ]


def str_to_date(value):
    """
    Parse string and make datetime.date of it
    """

    if not isinstance(value, str):
        raise TypeError("`value' must be inst of str")

    ret = None

    ret_attributes = set()
    re_res = DATE_EXPRESSION_C.match(value)

    if re_res:

        month = MONTHES.index(re_res.group('month')) + 1

        day = re_res.group('day')
        if len(day) == 2 and day[0] == ' ':
            ret_attributes.add('day_fixed')

        result = datetime.date(
            year=int(re_res.group('year')),
            month=month,
            day=int(day)
            )

        ret = result

    return ret, ret_attributes


def date_to_str(value, attrs=None):

    if not isinstance(value, datetime.date):
        raise TypeError("`value' must be inst of datetime.date")

    if attrs is None:
        attrs = set()

    month = MONTHES[value.month - 1]

    if value.day < 10 and 'day_fixed' in attrs:
        day_str = ' {}'.format(value.day)
    else:
        day_str = str(value.day)

    ret = '{day}-{month}-{year:04d}'.format(
        day=day_str,
        month=month,
        year=value.year
        )

    return ret


def str_to_datetime(value):
    """
    Parse string and make datetime.datetime of it
    """

    if not isinstance(value, str):
        raise TypeError("`value' must be inst of str")

    ret = None

    ret_attributes = set()
    re_res = DATETIME_EXPRESSION_C.match(value)

    if re_res:

        month = MONTHES.index(re_res.group('month')) + 1

        zonesign = '+'
        if re_res.group('zone_sign') == '-':
            zonesign = '-'

        tzinfo = wayround_org.utils.datetime_iso8601.gen_tz(
            int(re_res.group('zone_hours')),
            int(re_res.group('zone_minutes')),
            plus=zonesign == '+'
            )

        day = re_res.group('day')
        if len(day) == 2 and day[0] == ' ':
            ret_attributes.add('day_fixed')

        result = datetime.datetime(
            year=int(re_res.group('year')),
            month=month,
            day=int(day),
            hour=int(re_res.group('hour')),
            minute=int(re_res.group('minute')),
            second=int(re_res.group('second')),
            tzinfo=tzinfo
            )

        ret = result

    return ret, ret_attributes


def datetime_to_str(value, attrs=None):

    if not isinstance(value, datetime.datetime):
        raise TypeError("`value' must be inst of datetime.datetime")

    if not hasattr(value, 'tzinfo') or value.tzinfo is None:
        raise ValueError("supplied datetime must have time zone info")

    if attrs is None:
        attrs = set()

    month = MONTHES[value.month - 1]

    zone = wayround_org.utils.datetime_iso8601.format_tz(
        value.tzinfo,
        sep=False
        )

    if value.day < 10 and 'day_fixed' in attrs:
        day_str = ' {}'.format(value.day)
    else:
        day_str = str(value.day)

    ret = (
        '{day}-{month}-{year:04d} '
        '{hour:02d}:{minute:02d}:{second:02d} {zone}'
        ).format(
            day=day_str,
            month=month,
            year=value.year,
            hour=value.hour,
            minute=value.minute,
            second=value.second,
            zone=zone
            )

    return ret


def str_parse_test():
    """
    run tests if datetime string parser
    """

    tests = [
        '29-Jan-2016 13:36:08 +0200',
        '9-Jan-2016 13:36:08 +0200',
        ]

    for value in tests:

        print("test subject: {}".format(value))

        re_res = DATETIME_EXPRESSION_C.match(value)

        if re_res:

            for i in [
                    'day', 'month', 'year',
                    'hour', 'minute', 'second',
                    'zone_sign', 'zone_val',
                    'zone_hours', 'zone_minutes']:
                print("    {:20}: {}".format(i, re_res.group(i)))

            dt = str_to_datetime(value)[0]

            print(
                "    {:20}: {}".format(
                    'iso8601',
                    wayround_org.utils.datetime_iso8601.datetime_to_str(
                        dt
                        )
                    )
                )

            print(
                "    {:20}: {}".format(
                    'rendered',
                    datetime_to_str(
                        dt
                        )
                    )
                )
        else:
            print("    parse error")
    return




"""
This module is not even developed yet
"""

# TODO: work!

import json


def encrypt_data(data, password):

    ret = None

    wrapper = {
        'data': data
        }

    try:
        ret = json.dumps(wrapper)
    except:
        ret = None

    return ret


def decrypt_data(data, password):

    ret = None

    try:
        ret = json.loads(data)
    except:
        ret = None
    else:
        ret = ret['data']

    return ret




raise Exception("this module requires rework")
# TODO: this module requires deprications and deletions

import copy
import urllib
import re


def paths_by_path(path, last_el_type=None, path_is_absolute=None):
    '''
    Returns dict with three values:

       - path_str: path string
       - path_lst: list of path components
       - last_el_type: last element type: 'dir' or 'file'
       - path_is_absolute: True if path_str begins with '/'

    Input value can either be URI path part string ether list with path
    parts. If input type is list, then second parameter determines
    returning last_el_type.

    If input type is string, then last element type will be 'dir' if
    last char in string is '/', but it can be forced by last_el_type
    parameter.

    If input is list and second parameter is None, then returning
    last_el_type will be 'dir'. In both cases,

    if resulting last_el_type == 'dir', then path_str[-1:] == '/'.

    Parameter path_is_absolute can be used for manipulating '/' in
    begining of output path_str.

    examples:

    >>> urltools.paths_by_path('a/b/c/d')
    {'path_lst': ['a', 'b', 'c', 'd'],
     'path_str': '/a/b/c/d', 'last_el_type': 'file'}

    >>> urltools.paths_by_path('a/b/c/d/')
    {'path_lst': ['a', 'b', 'c', 'd'],
     'path_str': '/a/b/c/d/', 'last_el_type': 'dir'}

    >>> urltools.paths_by_path(['1','2','3','4'])
    {'path_lst': ['1', '2', '3', '4'],
     'path_str': '/1/2/3/4/', 'last_el_type': 'dir'}

    >>> urltools.paths_by_path(['1','2','3','4'], 'file')
    {'path_lst': ['1', '2', '3', '4'],
     'path_str': '/1/2/3/4', 'last_el_type': 'file'}

    '''
    if isinstance(path, list):
        path_str = '/'.join(path)
        path_lst = copy.copy(path)

        if last_el_type is None:
            last_el_type = 'dir'

    if isinstance(path, str):

        if last_el_type is None:
            if path[-1:] == '/':
                last_el_type = 'dir'
            else:
                last_el_type = 'file'

        if path_is_absolute is None:
            path_is_absolute = (path[0] == '/')

        t_path = path.strip('/')

        t_lst = t_path.split('/')

        path_str = '/' + t_path
        path_lst = []
        for i in t_lst:
            if i != '':
                path_lst.append(i)

    if last_el_type == 'dir':
        if path_str[-1:] != '/':
            path_str += '/'

    if path_is_absolute is None and path_str[0] != '/':
        path_str = '/' + path_str

    return {'path_str': path_str,
            'path_lst': copy.copy(path_lst),
            'last_el_type': last_el_type,
            'path_is_absolute': path_is_absolute}


def ischild(path1, path2):
    """
    Compairs two paths and returns True if path2 is child for path1
    """

    p1 = paths_by_path(path1)['path_lst']
    p2 = paths_by_path(path2)['path_lst']

    l1 = len(p1)
    l2 = len(p2)

    ret = True

    if l2 != (l1 + 1):
        ret = False

    if ret:
        for i in range(l1):
            if urllib.parse.unquote(p1[i]) != urllib.parse.unquote(p2[i]):
                ret = False
                break

    return ret

is_child = ischild


def parse_scheme_and_data(uri, ret_data=['scheme', 'data']):
    '''
    Parses URI and returns requested data.

    dici is returned if all ok. It will contain data relected by
    ret_data parameter.

    example:

       >>> parse_scheme_and_data('http://home.org:21/path')
       {'scheme': 'http', 'data': '//home.org:21/path'}
    '''

    ret = dict()

    re_res = re.match(r'(.*?):(.*)', uri)

    if re_res is not None:

        for i in ['scheme', 'data']:
            if i in ret_data:
                ret[i] = eval(i)

    else:
        ret = None

    return ret


def _parse_r(uri, part):
    s = parse_scheme_and_data(uri, [part])
    if s is None or s[part] is None:
        ret = None
    else:
        ret = s[part]
    return ret


def parse_scheme(uri):
    """Returns URI scheme part"""
    return _parse_r(uri, 'scheme')


def parse_data(uri):
    """Returns URI data part"""
    return _parse_r(uri, 'data')


def parse(
        url='http://login:password@example.net:80/some/path?with=parameters&an=d#anchor'
        ):
    """
    Parse URI and return None if error or dict structure accordingly
    to scheme name.

    This function, does not works with relative paths -- cases when
    there is no scheme present in URI -- None is returned.

    Currently, only http, https and ftp are allowed, and ftp treated
    as http.
    """
    scheme = parse_scheme(url)

    ret = None

    if scheme is None:
        ret = None
    else:
        re_res = re.match(r'(.*?):(.*)', url)

        if re_res is not None:
            scheme = re_res.group(1)
            data = re_res.group(2)

            if scheme in ['http', 'https', 'ftp']:
                ret = parse_all_data(data)
                ret['scheme'] = scheme
            else:
                ret = None

    return ret


def is_same_host(uri1, uri2):
    return is_same_site(uri1, uri2, False, False)


def is_same_site(
        uri1,
        uri2,
        not_if_scheme_not_eql=True,
        not_if_port_not_eql=True
        ):

    ret = True

    u1 = parse(uri1)
    u2 = parse(uri2)

    lst = ['host']

    if not_if_scheme_not_eql:
        lst.append('scheme')

    if not_if_port_not_eql:
        lst.append('port')

    for i in lst:
        if u1[i] != u2[i]:
            ret = False
            break

    return ret


def del_not_same_hosts(uri, lst):
    ret = list()

    for i in lst:
        if is_same_host(uri, i):
            ret.append(i)

    return ret


def del_not_same_sites(
        uri,
        lst,
        not_if_scheme_not_eql=True,
        not_if_port_not_eql=True
        ):
    """
    Takes URI. Takes URI list. And then, compares lst items to uri,
    forming new list to return.

    If host not same, lst item not goes to returning list.

    Additional parameters adds filtering by scheme and post: usualy,
    if scheme is not same - site assumed not same, but not all
    software thinking this way. E.g. noscript, assumes http://s.s and
    https://s.s the same site, but Apache httpd - does not.
    """

    ret = list()

    for i in lst:
        if is_same_site(uri, i, not_if_scheme_not_eql, not_if_port_not_eql):
            ret.append(i)

    return ret


def is_child_uri(
        uri1,
        uri2,
        not_if_scheme_not_eql=True,
        not_if_port_not_eql=True
        ):
    """
    Check is uri2 chiled to uri1
    """

    u1 = parse(uri1)
    u2 = parse(uri2)

    ret = True

    if u1 is None or u2 is None:
        ret = None

    else:
        if not is_same_site(
                uri1,
                uri2,
                not_if_scheme_not_eql,
                not_if_port_not_eql
                ):
            ret = None

        if not is_child(u1['path_lst'], u2['path_lst']):
            ret = False

    return ret


def del_not_child_uris(uri, lst):
    """
    Takes uri list, compares with uri and return list with uri childs
    """

    ret = list()

    for i in lst:
        if is_child_uri(uri, i):
            ret.append(i)

    return ret


def parse_parameters(data):
    """
    Parse URI parameter string to list and returns it

    Parameters can repeat

    examples:

       >>> parse_parameters('?a=1&b=2')
       [['a', '1'], ['b', '2']]

       >>> parse_parameters('a=1&b=2&a=3')
       [['a', '1'], ['b', '2'], ['a', '3']]

    """
    t_data = data.lstrip('?')

    data_splitted = t_data.split('&')

    para_list = []

    for i in data_splitted:
        para_splitted = i.split('=')
        if len(para_splitted) == 1:
            para_splitted.append(None)
        para_list.append(para_splitted)

    return copy.copy(para_list)


def parse_auth(data):
    """
    Parses URI authentication string to parts.

    Returns None if data == None or data == ''

    Else, returns dict with two values:

       - login: login part
       - password: password part

    examples:

       >>> parse_auth('123123@')
       {'login': '123123', 'password': None}

       >>> parse_auth('123123:@')
       {'login': '123123', 'password': ''}

       >>> parse_auth('123123:123@')
       {'login': '123123', 'password': '123'}

       >>> parse_auth(':123@')
       {'login': '', 'password': '123'}

       >>> parse_auth(':@')
       {'login': '', 'password': ''}

       >>> parse_auth(':')
       {'login': '', 'password': ''}

       >>> parse_auth('')
       None

       >>> parse_auth(None)
       None

    """
    if data is None or data == '':
        return None

    re_res = re.match(r'^(.*?)(:.*?)?@?$', data)

    if re_res is not None:
        login = re_res.group(1)
        password = None
        if re_res.group(2) is not None:
            password = re_res.group(2)[1:]
        else:
            password = None

        return {'login': login, 'password': password}
    else:
        return None


def parse_all_data(data):
    '''
    Can be deployed to strings like C{//agu:@wayround.org:21/1/2/3/4/?a#3}

    Such a string then will be parsed on several parts. Then thous
    parts will be worked out to dict like:

       >>> repr()
       {'anchor': '3',
        'auth': {'login': 'agu', 'password': ''},
        'host': 'wayround.org',
        'param_lst': [['a', None]],
        'path_lst': ['1', '2', '3', '4'],
        'path_str': '/1/2/3/4',
        'path_is_absolute': True,
        'last_el_type': 'file',
        'port': '21'}

    '''
    ret = dict(
        auth=None,  # or dict {'login':'', 'password':
        # ''}, where login must be string, and
        # password can be None
        host=None,
        port=None,
        path_str=None,
        path_lst=None,
        last_el_type=None,
        path_is_absolute=None,
        param_lst=None,
        anchor=None
        )

    re_res = re.match(r'^//(.*@)?(.*?)(:\d*)?(/.*?)?(\?.*?)?(\#.*)?$', data)

    if re_res is not None:
        # print 'auth: '+ repr(re_res.group(1))
        # print 'host: '+ repr(re_res.group(2))
        # print 'port: '+ repr(re_res.group(3))
        # print 'path: '+ repr(re_res.group(4))
        # print 'para: '+ repr(re_res.group(5))
        # print '  id: '+ repr(re_res.group(6))
        # print

        ret['auth'] = parse_auth(re_res.group(1))

        if re_res.group(2) is not None:
            ret['host'] = re_res.group(2)

        if re_res.group(3) is not None:
            ret['port'] = re_res.group(3)[1:]

        if re_res.group(4) is not None:
            t_p = paths_by_path(re_res.group(4))
            if t_p is None:
                ret['path_str'] = None
                ret['path_lst'] = None
                ret['last_el_type'] = None
                ret['path_is_absolute'] = None
            else:
                ret['path_str'] = t_p['path_str']
                ret['path_lst'] = t_p['path_lst']
                ret['last_el_type'] = t_p['last_el_type']
                ret['path_is_absolute'] = t_p['path_is_absolute']

        if re_res.group(5) is not None:
            ret['param_lst'] = parse_parameters(re_res.group(5))

        if re_res.group(6) is not None:
            ret['anchor'] = re_res.group(6)[1:]

        return ret
    return None


def combine_data(scheme='http',
                 auth={'login': 'anonymous',
                       'password': 'myemail'},
                 host='example.net',
                 port=80,
                 path='/',
                 parameters={},
                 anchor=''):
    '''
    Constructs URI string from parameters

    auth tuple can be complitly None or password can be None. auth
    login must allways be string or unicode, password too.

    port automaticly set to 80 or 443 if http or https scheme
    accordingly.

    parameters can be None or dicrionary like {'name':'value'}
    '''

    auth_str = ''
    port_str = ''
    use_port = False
    param_str = ''
    use_param = False
    anchor_str = ''
    use_anchor = False

    if auth is not None:
        auth_str = auth[0]
        if auth[1] is not None:
            auth_str += ':' + auth[1]
        auth_str += '@'

    if scheme == 'http' and port != 80:
        use_port = True

    if scheme == 'https' and port != 443:
        use_port = True

    if use_port:
        port_str = ':' + str(port)

    if parameters is not None and len(parameters.keys()) > 0:
        use_param = True

    if isinstance(path, list):
        path = '/'.join(path)

    if use_param:
        param_str = '?'
        param_lst = []
        for i in parameters.keys():
            p2 = parameters[i]

            if isinstance(p2, (int, float)):
                p2 = str(p2)

            if isinstance(p2, bytes):
                p2.decode('utf-8')

            param_lst.append(str(i) + '=' + p2)

        param_str += '&'.join(param_lst)

    if anchor is not None and anchor != '':
        use_anchor = True

    if use_anchor:
        anchor_str = '#' + anchor

    res = \
        scheme \
        + '://' \
        + auth_str \
        + host \
        + port_str \
        + path \
        + param_str \
        + anchor_str
    return res
